/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file station_area_func.cpp Implementation of the StationArea functions. */





#include "station_area.h"
#include "station_area_func.h"

#include "../command_func.h"

#include "debug.h"




//StationArea::StationArea (
//) {

//	EmptyStationArea (this);
//}





void EmptyStationArea (
	StationArea *station_area
) {

	station_area->left = station_area->top = station_area->right = station_area->bottom = 0;
}





/**
 * Determines whether a given point (x, y) is within a certain distance of
 * the station rectangle.
 * @note x and y are in Tile coordinates
 * @param x X coordinate
 * @param y Y coordinate
 * @param distance The maximum distance a point may have (L1 norm)
 * @return true if the point is within distance tiles of the station rectangle
 */

bool PtInExtendedRect (
	StationArea *station_area, int x, int y, int distance
) {

	return station_area->left - distance <= x && x <= station_area->right + distance &&
			station_area->top - distance <= y && y <= station_area->bottom + distance;
}





bool IsStationAreaEmpty (
	const StationArea *station_area
) {

	return station_area->left == 0 || station_area->left > station_area->right || station_area->top > station_area->bottom;
}





CommandCost BeforeAddTile (
	StationArea *station_area, TileIndex tile, StationAreaMode mode
) {

	int x = TileX(tile);
	int y = TileY(tile);

	if (IsStationAreaEmpty (station_area)) {

		/* we are adding the first station tile */
		if (mode != ADD_TEST) {

			station_area->left = station_area->right = x;
			station_area->top = station_area->bottom = y;
		}

	} else if (!PtInExtendedRect (station_area, x, y)) {

		/* current rect is not empty and new point is outside this rect
		 * make new spread-out rectangle */
		Rect new_rect = {std::min(x, station_area->left), std::min(y, station_area->top), std::max(x, station_area->right), std::max(y, station_area->bottom)};

		/* check new rect dimensions against preset max */
		int w = new_rect.right - new_rect.left + 1;
		int h = new_rect.bottom - new_rect.top + 1;

		if (mode != ADD_FORCE && (w > _settings_game.station.station_spread || h > _settings_game.station.station_spread)) {

			assert(mode != ADD_TRY);
			return_cmd_error(STR_ERROR_STATION_TOO_SPREAD_OUT);
		}

		/* spread-out ok, return true */
		if (mode != ADD_TEST) {

			/* we should update the station rect */
			CopyStationArea (station_area, &new_rect);
		}

	} else {

		; // new point is inside the rect, we don't need to do anything
	}

	return CommandCost();
}





CommandCost BeforeAddRect (
	StationArea *station_area, TileIndex tile, int w, int h, StationAreaMode mode
) {

	if (mode == ADD_FORCE || (w <= _settings_game.station.station_spread && h <= _settings_game.station.station_spread)) {

		/* Important when the old rect is completely inside the new rect, resp. the old one was empty. */
		CommandCost ret = BeforeAddTile (station_area, tile, mode);

		if (ret.Succeeded()) ret = BeforeAddTile (station_area, TILE_ADDXY(tile, w - 1, h - 1), mode);

		return ret;
	}

	return CommandCost();
}





/**
 * Check whether station tiles of the given station id exist in the given rectangle
 * @param st_id    Station ID to look for in the rectangle
 * @param left_a   Minimal tile X edge of the rectangle
 * @param top_a    Minimal tile Y edge of the rectangle
 * @param right_a  Maximal tile X edge of the rectangle (inclusive)
 * @param bottom_a Maximal tile Y edge of the rectangle (inclusive)
 * @return \c true if a station tile with the given \a st_id exists in the rectangle, \c false otherwise
 */

bool ScanForStationTiles (
	StationArea *station_area, StationID st_id, int left_a, int top_a, int right_a, int bottom_a
) {

	TileArea ta(TileXY(left_a, top_a), TileXY(right_a, bottom_a));
	for (TileIndex tile : ta) {
		if (IsTileType(tile, MP_STATION) && GetStationIndex(tile) == st_id) return true;
	}

	return false;
}





bool AfterRemoveTile (
	StationArea *station_area, BaseStation *station, TileIndex tile
) {

	int x = TileX(tile);
	int y = TileY(tile);

	/* look if removed tile was on the bounding rect edge
	 * and try to reduce the rect by this edge
	 * do it until we have empty rect or nothing to do */
	for (;;) {
		/* check if removed tile is on rect edge */
		bool left_edge = (x == station_area->left);
		bool right_edge = (x == station_area->right);
		bool top_edge = (y == station_area->top);
		bool bottom_edge = (y == station_area->bottom);

		/* can we reduce the rect in either direction? */
		bool reduce_x = ((left_edge
						  || right_edge
						 )
						 && !ScanForStationTiles (station_area, station->index,
												  x, station_area->top, x, station_area->bottom)
						)
		;

		bool reduce_y = ((top_edge
						  || bottom_edge
						 )
						 && !ScanForStationTiles (station_area, station->index,
												  station_area->left, y, station_area->right, y)
						)
		;

		if (!(reduce_x || reduce_y)) break; // nothing to do (can't reduce)

		if (reduce_x) {
			/* reduce horizontally */
			if (left_edge) {
				/* move left edge right */
				station_area->left = x = x + 1;
			} else {
				/* move right edge left */
				station_area->right = x = x - 1;
			}
		}
		if (reduce_y) {
			/* reduce vertically */
			if (top_edge) {
				/* move top edge down */
				station_area->top = y = y + 1;
			} else {
				/* move bottom edge up */
				station_area->bottom = y = y - 1;
			}
		}

		if (station_area->left > station_area->right || station_area->top > station_area->bottom) {

			/* can't continue, if the remaining rectangle is empty */
			EmptyStationArea (station_area);

			return true; // empty remaining rect
		}
	}

	return false; // non-empty remaining rect
}





bool AfterRemoveRect (
	StationArea *station_area, BaseStation *station, TileArea ta
) {

	assert (PtInExtendedRect (station_area, TileX(ta.tile), TileY(ta.tile)));
	assert (PtInExtendedRect (station_area, TileX(ta.tile) + ta.w - 1, TileY(ta.tile) + ta.h - 1));

	bool empty = AfterRemoveTile (station_area, station, ta.tile);
	if (ta.w != 1 || ta.h != 1) empty = empty || AfterRemoveTile (station_area, station, TILE_ADDXY(ta.tile, ta.w - 1, ta.h - 1));
	return empty;
}





//StationArea& StationArea::operator = (const Rect &src)

StationArea& CopyStationArea (
	StationArea *dest_area, Rect *src_rect
) {

	dest_area->left   = src_rect->left;
	dest_area->top    = src_rect->top;
	dest_area->right  = src_rect->right;
	dest_area->bottom = src_rect->bottom;

	return *dest_area;
}
