/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file station_func.h Functions related to stations. */



#ifndef STATION_FUNC_H
#define STATION_FUNC_H



#include "../economy/economy_func.h"
#include "../industry/industry_type.h"

#include "../infrastructure/rail_type.h"
#include "../infrastructure/road_type.h"
#include "../infrastructure/road.h"

#include "../linkgraph/linkgraph_type.h"

#include "../newgrf/newgrf_airport.h"
#include "../newgrf/newgrf_station.h"

#include "../vehicle/vehicle_type.h"
#include "../vehicle/vehicle_base.h"
#include "../vehicle/train.h"

#include "../sprite.h"



enum StationNaming {
	STATIONNAMING_RAIL,
	STATIONNAMING_ROAD,
	STATIONNAMING_AIRPORT,
	STATIONNAMING_OILRIG,
	STATIONNAMING_DOCK,
	STATIONNAMING_HELIPORT,
};





static const TileIndexDiffC _dock_tileoffs_chkaround[] = {
	{-1,  0},
	{ 0,  0},
	{ 0,  0},
	{ 0, -1}
};
static const byte _dock_w_chk[4] = { 2, 1, 2, 1 };
static const byte _dock_h_chk[4] = { 1, 2, 1, 2 };





void ModifyStationRatingAround(TileIndex tile, Owner owner, int amount, uint radius);

void ShowStationViewWindow(StationID station);
void UpdateAllStationVirtCoords();
void ClearAllStationCachedNames();

CargoArray GetProductionAroundTiles(TileIndex tile, int w, int h, int rad);
CargoArray GetAcceptanceAroundTiles(TileIndex tile, int w, int h, int rad, CargoTypes *always_accepted = nullptr);

void UpdateStationAcceptance(Station *st, bool show_msg);

const DrawTileSprites *GetStationTileLayout(StationType st, byte gfx);
void StationPickerDrawSprite(int x, int y, StationType st, RailType railtype, RoadType roadtype, int image);

bool HasStationInUse(StationID station, bool include_company, CompanyID company);

uint MoveGoodsToStation(CargoID type, uint amount, SourceType source_type, SourceID source_id, const StationList *all_stations, Owner exclusivity = INVALID_OWNER);
void DeleteOilRig(TileIndex t);
void UpdateStationDockingTiles(Station *st);
void RemoveDockingTile(TileIndex t);
void ClearDockingTilesCheckingNeighbours(TileIndex tile);
bool IsValidDockingDirectionForDock(TileIndex t, DiagDirection d);

/* Check if a rail station tile is traversable. */
bool IsStationTileBlocked(TileIndex tile);

bool CanStationTileHavePylons(TileIndex tile);
bool CanStationTileHaveWires(TileIndex tile);

void UpdateAirportsNoise();

bool SplitGroundSpriteForOverlay(const TileInfo *ti, SpriteID *ground, RailTrackOffset *overlay_offset);

void IncreaseStats(Station *st, const Vehicle *v, StationID next_station_id, TimeTick time);
void IncreaseStats(Station *st, CargoID cargo, StationID next_station_id, int64 capacity, int64 usage, TimeTick time, EdgeUpdateMode mode);
void RerouteCargo(Station *st, CargoID c, StationID avoid, StationID avoid2);

CommandCost CheckFlatLandRailStation (
	TileArea tile_area, ProcessCmdFlag flags, Axis axis, StationID *station, RailType rt,
	std::vector<Train *> &affected_vehicles, StationClassID spec_class, byte spec_index, byte plat_len, byte numtracks
);
CommandCost FindJoiningStation(StationID existing_station, StationID station_to_join, bool adjacent, TileArea ta, Station **st);
CommandCost BuildStationPart(Station **st, ProcessCmdFlag flags, bool reuse, TileArea area, StationNaming name_class);
void GetStationLayout(byte *layout, uint numtracks, uint plat_len, const StationSpec *statspec);
void FreeTrainReservation(Train *v);
void RestoreTrainReservation(Train *v);

template <class T>
CommandCost RemoveFromRailBaseStation(TileArea ta, std::vector<T *> &affected_stations, ProcessCmdFlag flags, Money removal_cost, bool keep_rail);
CommandCost CheckFlatLandRoadStop(TileArea tile_area, ProcessCmdFlag flags, uint invalid_dirs, bool is_drive_through, bool is_truck_stop, Axis axis, StationID *station, RoadType rt);
CommandCost FindJoiningRoadStop(StationID existing_stop, StationID station_to_join, bool adjacent, TileArea ta, Station **st);
CommandCost RemoveRoadStop(TileIndex tile, ProcessCmdFlag flags);
RoadStop **FindRoadStopSpot(bool truck_station, Station *st);
CommandCost CheckFlatLandAirport(AirportTileTableIterator tile_iter, ProcessCmdFlag flags);
Town *AirportGetNearestTown(const AirportSpec *as, const TileIterator &it, uint &mindist);
uint8 GetAirportNoiseLevelForDistance(const AirportSpec *as, uint distance);
bool IsUniqueStationName(const std::string &name);





/**
 * Calculates the maintenance cost of a number of station tiles.
 * @param num Number of station tiles.
 * @return Total cost.
 */
static inline Money StationMaintenanceCost(uint32 num)
{
	return (_price[PR_INFRASTRUCTURE_STATION] * num * (1 + IntSqrt(num))) >> 7; // 7 bits scaling.
}



Money AirportMaintenanceCost(Owner owner);



#endif /* STATION_FUNC_H */
