/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file map_func.h Functions related to maps. */



#ifndef MAP_FUNC_H
#define MAP_FUNC_H

#include "map_type.h"

#include "../core/math_func.hpp"
#include "../tile_type.h"
#include "../direction_func.h"





/**
 * 'Wraps' the given tile to it is within the map. It does
 * this by masking the 'high' bits of.
 * @param x the tile to 'wrap'
 */

#define TILE_MASK(x) ((x) & _map_tile_mask)

/**
 * An offset value between to tiles.
 *
 * This value is used for the difference between
 * two tiles. It can be added to a tileindex to get
 * the resulting tileindex of the start tile applied
 * with this saved difference.
 *
 * @see TileDiffXY(int, int)
 */
typedef int32 TileIndexDiff;

extern uint _map_tile_mask;

/**
 * Pointer to the tile-array.
 *
 * This variable points to the tile-array which contains the tiles of
 * the map.
 */
extern Tile *_m;

/**
 * Pointer to the extended tile-array.
 *
 * This variable points to the extended tile-array which contains the tiles
 * of the map.
 */
extern TileExtended *_me;

/**
 * Pointer to the tile-array cache.
 *
 * This variable points to the tile-array cache which contains frequently used tiles
 * of the map.
 */
extern TileCache *_mc;

void AllocateMap(uint size_x, uint size_y);

uint MapLogX();

uint MapLogY();

uint32 MapSizeX ();

uint32 MapSizeY ();

uint32 MapSize ();

uint32 MapMaxX ();

uint32 MapMaxY ();

uint ScaleByMapSize(uint n);

uint ScaleByMapSize1D(uint n);

TileIndex TileXY(uint x, uint y);

TileIndexDiff TileDiffXY(int x, int y);

TileIndex TileVirtXY(uint x, uint y);

uint32 TileX (TileIndex tile);

uint32 TileY (TileIndex tile);

TileIndexDiff ToTileIndexDiff(TileIndexDiffC tidc);

#ifndef _DEBUG
	/**
	 * Adds to tiles together.
	 *
	 * @param x One tile
	 * @param y Another tile to add
	 * @return The resulting tile(index)
	 */
#	define TILE_ADD(x, y) ((x) + (y))
#else
	extern TileIndex TileAdd(TileIndex tile, TileIndexDiff add,
		const char *exp, const char *file, int line);
#	define TILE_ADD(x, y) (TileAdd((x), (y), #x " + " #y, __FILE__, __LINE__))
#endif

/**
 * Adds a given offset to a tile.
 *
 * @param tile The tile to add an offset on it
 * @param x The x offset to add to the tile
 * @param y The y offset to add to the tile
 */
#define TILE_ADDXY(tile, x, y) TILE_ADD(tile, TileDiffXY(x, y))

TileIndex TileAddWrap(TileIndex tile, int addx, int addy);

TileIndexDiffC TileIndexDiffCByDiagDir(DiagDirection dir);

TileIndexDiffC TileIndexDiffCByDir(Direction dir);

TileIndex AddTileIndexDiffCWrap(TileIndex tile, TileIndexDiffC diff);

TileIndexDiffC TileIndexToTileIndexDiffC (TileIndex tile_a, TileIndex tile_b);

/* Functions to calculate distances */
uint DistanceManhattan(TileIndex, TileIndex); ///< also known as L1-Norm. Is the shortest distance one could go over diagonal tracks (or roads)
uint DistanceSquare(TileIndex, TileIndex); ///< euclidian- or L2-Norm squared
uint DistanceMax(TileIndex, TileIndex); ///< also known as L-Infinity-Norm
uint DistanceMaxPlusManhattan(TileIndex, TileIndex); ///< Max + Manhattan
uint DistanceFromEdge(TileIndex); ///< shortest distance from any edge of the map
uint DistanceFromEdgeDir(TileIndex, DiagDirection); ///< distance from the map edge in given direction

TileIndexDiff TileOffsByDiagDir(DiagDirection dir);

TileIndexDiff TileOffsByDir(Direction dir);

TileIndex TileAddByDir(TileIndex tile, Direction dir);

TileIndex TileAddByDiagDir(TileIndex tile, DiagDirection dir);

DiagDirection DiagdirBetweenTiles(TileIndex tile_from, TileIndex tile_to);

typedef bool TestTileOnSearchProc(TileIndex tile, void *user_data);

bool CircularTileSearch(TileIndex *tile, uint size, TestTileOnSearchProc proc, void *user_data);
bool CircularTileSearch(TileIndex *tile, uint radius, uint w, uint h, TestTileOnSearchProc proc, void *user_data);

TileIndex RandomTileSeed(uint32 r);

#define RandomTile() RandomTileSeed(Random())

uint GetClosestWaterDistance(TileIndex tile, bool water);





#endif /* MAP_FUNC_H */
