/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file order.cpp Handling of order functions. */





#include "../stdafx.h"

#include "order_type.h"
#include "order.h"

#include "../core/pool_func.hpp"
#include "../gui/window_func.h"
#include "../facilities/station_base.h"

#include "../debug.h"
#include "../safeguards.h"





OrderPool _order_pool("Order");
INSTANTIATE_POOL_METHODS(Order)





/** Clean everything up. */
Order::~Order()
{
	if (CleaningPool()) return;

	/* We can visit oil rigs and buoys that are not our own. They will be shown in
	 * the list of stations. So, we need to invalidate that window if needed. */
	if (this->IsType(OT_GOTO_STATION) || this->IsType(OT_GOTO_WAYPOINT)) {
		BaseStation *bs = BaseStation::GetIfValid(this->dest);
		if (bs != nullptr && bs->owner == OWNER_NONE) InvalidateWindowClassesData(WC_STATION_LIST, 0);
	}
}





/**
 * 'Free' the order
 * @note ONLY use on "current_order" vehicle orders!
 */

void Order::Free(
) {

	this->type  = OT_NOTHING;
	this->flags = 0;
	this->dest  = 0;
	this->next  = nullptr;
}





/**
 * Makes this order a Go To Station order.
 * @param destination the station to go to.
 */

void Order::MakeGoToStation (
	StationID destination
) {

	this->type  = OT_GOTO_STATION;
	this->flags = 0;
	this->dest  = destination;
}





/**
 * Makes this order a Go To Depot order.
 * @param destination   the depot to go to.
 * @param order         is this order a 'default' order, or an overridden vehicle order?
 * @param non_stop_type how to get to the depot?
 * @param action        what to do in the depot?
 * @param cargo         the cargo type to change to.
 */

void Order::MakeGoToDepot (
	DepotID destination, OrderDepotTypeFlags order, OrderNonStopFlags non_stop_type,
	OrderDepotActionFlags action, CargoID cargo
) {

	this->type               = OT_GOTO_DEPOT;
	this->SetDepotOrderType  (order);
	this->SetDepotActionType (action);
	this->SetNonStopType     (non_stop_type);
	this->dest               = destination;
	this->SetRefit           (cargo);
}





/**
 * Makes this order a Go To Waypoint order.
 * @param destination the waypoint to go to.
 */

void Order::MakeGoToWaypoint (
	StationID destination
) {

	this->type  = OT_GOTO_WAYPOINT;
	this->flags = 0;
	this->dest  = destination;
}





/**
 * Makes this order an conditional order.
 * @param order the order to jump to.
 */

void Order::MakeConditional (
	VehicleOrderID order
) {

	this->type  = OT_CONDITIONAL;
	this->flags = order;
	this->dest  = 0;
}





/**
 * Make this depot/station order also a refit order.
 * @param cargo   the cargo type to change to.
 * @pre IsType(OT_GOTO_DEPOT) || IsType(OT_GOTO_STATION).
 */
void Order::SetRefit(CargoID cargo)
{
	this->refit_cargo = cargo;
}





/**
 * Does this order have the same type, flags and destination?
 * @param other the second order to compare to.
 * @return true if the type, flags and destination match.
 */
bool Order::Equals(const Order &other) const
{
	/* In case of go to nearest depot orders we need "only" compare the flags
	 * with the other and not the nearest depot order bit or the actual
	 * destination because those get clear/filled in during the order
	 * evaluation. If we do not do this the order will continuously be seen as
	 * a different order and it will try to find a "nearest depot" every tick. */
	if ((this->IsType(OT_GOTO_DEPOT) && this->type == other.type) &&
			((this->GetDepotActionType() & ODATFB_NEAREST_DEPOT) != 0 ||
			 (other.GetDepotActionType() & ODATFB_NEAREST_DEPOT) != 0)) {
		return this->GetDepotOrderType() == other.GetDepotOrderType() &&
				(this->GetDepotActionType() & ~ODATFB_NEAREST_DEPOT) == (other.GetDepotActionType() & ~ODATFB_NEAREST_DEPOT);
	}

	return this->type == other.type && this->flags == other.flags && this->dest == other.dest;
}





/**
 * Pack this order into a 32 bits integer, or actually only
 * the type, flags and destination.
 * @return the packed representation.
 * @note unpacking is done in the constructor.
 */
uint32 Order::Pack() const
{
	return this->dest << 16 | this->flags << 8 | this->type;
}





/**
 * Pack this order into a 16 bits integer as close to the TTD
 * representation as possible.
 * @return the TTD-like packed representation.
 */
uint16 Order::MapOldOrder() const
{
	uint16 order = this->GetType();
	switch (this->type) {
		case OT_GOTO_STATION:
			if (this->GetUnloadType() & OUFB_UNLOAD) SetBit(order, 5);
			if (this->GetLoadType() & OLFB_FULL_LOAD) SetBit(order, 6);
			if (this->GetNonStopType() & ONSF_NO_STOP_AT_INTERMEDIATE_STATIONS) SetBit(order, 7);
			order |= GB(this->dest, 0, 8) << 8;
			break;
		case OT_GOTO_DEPOT:
			if (!(this->GetDepotOrderType() & ODTFB_PART_OF_ORDERS)) SetBit(order, 6);
			SetBit(order, 7);
			order |= GB(this->dest, 0, 8) << 8;
			break;
		case OT_LOADING:
			if (this->GetLoadType() & OLFB_FULL_LOAD) SetBit(order, 6);
			break;
	}
	return order;
}





/**
 * Create an order based on a packed representation of that order.
 * @param packed the packed representation.
 */
Order::Order(uint32 packed)
{
	this->type    = (OrderType)GB(packed,  0,  8);
	this->flags   = GB(packed,  8,  8);
	this->dest    = GB(packed, 16, 16);
	this->next    = nullptr;
	this->refit_cargo   = CT_NO_REFIT;
	this->wait_time     = 0;
	this->travel_time   = 0;
	this->max_speed     = UINT16_MAX;
}





/**
 *
 * Assign data to an order (from another order)
 *   This function makes sure that the index is maintained correctly
 * @param other the data to copy (except next pointer).
 *
 */
void Order::AssignOrder(const Order &other)
{
	this->type  = other.type;
	this->flags = other.flags;
	this->dest  = other.dest;

	this->refit_cargo   = other.refit_cargo;

	this->wait_time   = other.wait_time;
	this->travel_time = other.travel_time;
	this->max_speed   = other.max_speed;
}





/**
 * Check if vehicle can unload/load at station.
 */

bool CanLoadOrUnload (
	const Order *order
) {

	return order->IsType (OT_GOTO_STATION)
			&& (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION) == 0
			&& (
				(order->GetLoadType() & OLFB_NO_LOAD) == 0
				|| (order->GetUnloadType() & OUFB_NO_UNLOAD) == 0
			)
	;
}





/**
 * A vehicle can leave the current station with cargo if:
 * 1. it can load cargo here OR
 * 2a. it could leave the last station with cargo AND
 * 2b. it doesn't have to unload all cargo here.
 */

bool CanLeaveWithCargo (
	const Order *order, bool has_cargo
) {

	return (order->GetLoadType() & OLFB_NO_LOAD) == 0
			|| (has_cargo
				&& (order->GetUnloadType() & (OUFB_UNLOAD | OUFB_TRANSFER)) == 0
			)
	;
}





	/**
	 * Check whether this order is of the given type.
	 * @param type the type to check against.
	 * @return true if the order matches.
	 */
	bool Order::IsType(OrderType type) const { return this->GetType() == type; }





	/**
	 * Get the type of order of this order.
	 * @return the order type.
	 */
	OrderType Order::GetType() const { return (OrderType)GB(this->type, 0, 4); }





	/**
	 * Is this a 'goto' order with a real destination?
	 * @return True if the type is either #OT_GOTO_WAYPOINT, #OT_GOTO_DEPOT or #OT_GOTO_STATION.
	 */
	bool Order::IsGotoOrder() const
	{
		return IsType(OT_GOTO_WAYPOINT) || IsType(OT_GOTO_DEPOT) || IsType(OT_GOTO_STATION);
	}





	/**
	 * Gets the destination of this order.
	 * @pre IsType(OT_GOTO_WAYPOINT) || IsType(OT_GOTO_DEPOT) || IsType(OT_GOTO_STATION).
	 * @return the destination of the order.
	 */
	DestinationID Order::GetDestination() const { return this->dest; }





	/**
	 * Sets the destination of this order.
	 * @param destination the new destination of the order.
	 * @pre IsType(OT_GOTO_WAYPOINT) || IsType(OT_GOTO_DEPOT) || IsType(OT_GOTO_STATION).
	 */
	void Order::SetDestination(DestinationID destination) { this->dest = destination; }





	/**
	 * Is this order a refit order.
	 * @pre IsType(OT_GOTO_DEPOT) || IsType(OT_GOTO_STATION)
	 * @return true if a refit should happen.
	 */
	bool Order::IsRefit() const { return this->refit_cargo < NUM_CARGO || this->refit_cargo == CT_AUTO_REFIT; }





	/**
	 * Is this order a auto-refit order.
	 * @pre IsType(OT_GOTO_DEPOT) || IsType(OT_GOTO_STATION)
	 * @return true if a auto-refit should happen.
	 */
	bool Order::IsAutoRefit() const { return this->refit_cargo == CT_AUTO_REFIT; }





	/**
	 * Get the cargo to to refit to.
	 * @pre IsType(OT_GOTO_DEPOT) || IsType(OT_GOTO_STATION)
	 * @return the cargo type.
	 */
	CargoID Order::GetRefitCargo() const { return this->refit_cargo; }





	/** How must the consist be loaded? */
	OrderLoadFlags Order::GetLoadType() const { return (OrderLoadFlags)GB(this->flags, 4, 3); }

	/** How must the consist be unloaded? */
	OrderUnloadFlags Order::GetUnloadType() const { return (OrderUnloadFlags)GB(this->flags, 0, 3); }

	/** At which stations must we stop? */
	OrderNonStopFlags Order::GetNonStopType() const { return (OrderNonStopFlags)GB(this->type, 6, 2); }

	/** Where must we stop at the platform? */
	OrderStopLocation Order::GetStopLocation() const { return (OrderStopLocation)GB(this->type, 4, 2); }

	/** What caused us going to the depot? */
	OrderDepotTypeFlags Order::GetDepotOrderType() const { return (OrderDepotTypeFlags)GB(this->flags, 0, 3); }

	/** What are we going to do when in the depot. */
	OrderDepotActionFlags Order::GetDepotActionType() const { return (OrderDepotActionFlags)GB(this->flags, 4, 3); }

	/** What waypoint flags? */
	OrderWaypointFlags Order::GetWaypointFlags() const { return (OrderWaypointFlags)GB(this->flags, 0, 3); }

	/** What variable do we have to compare? */
	OrderConditionVariable Order::GetConditionVariable() const { return (OrderConditionVariable)GB(this->dest, 11, 5); }

	/** What is the comparator to use? */
	OrderConditionComparator Order::GetConditionComparator() const { return (OrderConditionComparator)GB(this->type, 5, 3); }

	/** Get the order to skip to. */
	VehicleOrderID Order::GetConditionSkipToOrder() const { return this->flags; }

	/** Get the value to base the skip on. */
	uint16 Order::GetConditionValue() const { return GB(this->dest, 0, 11); }



	/** Set how the consist must be loaded. */
	void Order::SetLoadType(OrderLoadFlags load_type) { SB(this->flags, 4, 3, load_type); }

	/** Set how the consist must be unloaded. */
	void Order::SetUnloadType(OrderUnloadFlags unload_type) { SB(this->flags, 0, 3, unload_type); }

	/** Set whether we must stop at stations or not. */
	void Order::SetNonStopType(OrderNonStopFlags non_stop_type) { SB(this->type, 6, 2, non_stop_type); }

	/** Set where we must stop at the platform. */
	void Order::SetStopLocation(OrderStopLocation stop_location) { SB(this->type, 4, 2, stop_location); }

	/** Set the cause to go to the depot. */
	void Order::SetDepotOrderType(OrderDepotTypeFlags depot_order_type) { SB(this->flags, 0, 3, depot_order_type); }

	/** Set what we are going to do in the depot. */
	void Order::SetDepotActionType(OrderDepotActionFlags depot_service_type) { SB(this->flags, 4, 3, depot_service_type); }

	/** Set waypoint flags. */
	void Order::SetWaypointFlags(OrderWaypointFlags waypoint_flags) { SB(this->flags, 0, 3, waypoint_flags); }

	/** Set variable we have to compare. */
	void Order::SetConditionVariable(OrderConditionVariable condition_variable) { SB(this->dest, 11, 5, condition_variable); }

	/** Set the comparator to use. */
	void Order::SetConditionComparator(OrderConditionComparator condition_comparator) { SB(this->type, 5, 3, condition_comparator); }

	/** Get the order to skip to. */
	void Order::SetConditionSkipToOrder(VehicleOrderID order_id) { this->flags = order_id; }

	/** Set the value to base the skip on. */
	void Order::SetConditionValue(uint16 value) { SB(this->dest, 0, 11, value); }
