/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file order_base.h Base class for orders. */





#ifndef ORDER_H
#define ORDER_H





#include "order_type.h"

#include "../core/bitmath_func.hpp"

#include "../cargo/cargo_type.h"

#include "../facilities/depot_type.h"
#include "../facilities/station_type.h"

#include "../saveload/saveload.h"
#include "../vehicle/vehicle_type.h"





/* If you change this, keep in mind that it is saved on 3 places:
 * - Load_ORDR, all the global orders
 * - Vehicle -> current_order
 * - REF_ORDER (all REFs are currently limited to 16 bits!!)
 */

struct Order : OrderPool::PoolItem<&_order_pool> {

	friend struct  VEHSChunkHandler;                     ///< Loading of ancient vehicles.
	friend         SaveLoadTable GetOrderDescription();  ///< Saving and loading of orders.
	friend class   SlVehicleCommon;
	friend class   SlVehicleDisaster;

	uint8           type;                 ///< The type of order + non-stop flags
	uint8           flags;                ///< Load/unload types, depot order/action types.
	int64           order_flags;          ///< Flags specific to the order itself.
	DestinationID   dest;                 ///< The destination of the order.

	CargoID         refit_cargo;          ///< Refit CargoID

	TimeTick        timed_arrival;        ///< The time of actual arrival at the destination.
	TimeTick        timed_departure;      ///< The time of actual departure.
	TimeTick        scheduled_arrival;    ///< The time of scheduled arrival.
	TimeTick        scheduled_departure;  ///< The time of scheduled departure.
	TimeTick        wait_time;            ///< How long in ticks to wait at the destination.
	TimeTick        travel_time;          ///< How long in ticks the journey to this destination should take.
	uint16          max_speed;            ///< How fast the vehicle may go on the way to the destination.

	Order          *next;                 ///< Pointer to next order. If nullptr, end of list

	Order() : flags(0), refit_cargo(CT_NO_REFIT),
			  timed_arrival (0), timed_departure(0), scheduled_arrival(0), scheduled_departure(0),
			  wait_time(0), max_speed(UINT16_MAX)
	{}
	~Order();

	Order(uint32 packed);

	bool IsType(OrderType type) const;
	OrderType GetType() const;

	void Free();

	void MakeGoToStation(StationID destination);
	void MakeGoToDepot(DepotID destination, OrderDepotTypeFlags order, OrderNonStopFlags non_stop_type = ONSF_NO_STOP_AT_INTERMEDIATE_STATIONS, OrderDepotActionFlags action = ODATF_SERVICE_ONLY, CargoID cargo = CT_NO_REFIT);
	void MakeGoToWaypoint(StationID destination);
	void MakeLoading(bool ordered);
	void MakeLeaveStation();
	void MakeDummy();
	void MakeConditional(VehicleOrderID order);
	void MakeWaiting();

	bool IsGotoOrder() const;
	DestinationID GetDestination() const;
	void SetDestination(DestinationID destination);
	bool IsRefit() const;
	bool IsAutoRefit() const;

	CargoID GetRefitCargo() const;

	void SetRefit(CargoID cargo);
	OrderLoadFlags GetLoadType() const;
	OrderUnloadFlags GetUnloadType() const;
	OrderNonStopFlags GetNonStopType() const;
	OrderStopLocation GetStopLocation() const;
	OrderDepotTypeFlags GetDepotOrderType() const;
	OrderDepotActionFlags GetDepotActionType() const;
	OrderWaypointFlags GetWaypointFlags() const;
	OrderConditionVariable GetConditionVariable() const;
	OrderConditionComparator GetConditionComparator() const;
	VehicleOrderID GetConditionSkipToOrder() const;
	uint16 GetConditionValue() const;

	void SetLoadType(OrderLoadFlags load_type);
	void SetUnloadType(OrderUnloadFlags unload_type);
	void SetNonStopType(OrderNonStopFlags non_stop_type);
	void SetStopLocation(OrderStopLocation stop_location);
	void SetDepotOrderType(OrderDepotTypeFlags depot_order_type);
	void SetDepotActionType(OrderDepotActionFlags depot_service_type);
	void SetWaypointFlags(OrderWaypointFlags waypoint_flags);
	void SetConditionVariable(OrderConditionVariable condition_variable);
	void SetConditionComparator(OrderConditionComparator condition_comparator);
	void SetConditionSkipToOrder(VehicleOrderID order_id);
	void SetConditionValue(uint16 value);

	void AssignOrder(const Order &other);
	bool Equals(const Order &other) const;

	uint32 Pack() const;
	uint16 MapOldOrder() const;
};



#endif /* ORDER_H */
