/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file order_cmd.cpp Handling of orders. */



/*

split file into separate ones for functions for order, orderlist, cmd functions, order logic

*/




#include "../stdafx.h"

#include "order_func_local.h"
#include "order_func_global.h"
#include "order_logic.h"
#include "order_backup.h"
#include "vehicle_orders.h"

#include "timetable.h"

#include "../core/pool_func.hpp"
#include "../core/random_func.hpp"

#include "../economy/company_base.h"
#include "../economy/company_func.h"

#include "../facilities/depot_base.h"
#include "../facilities/station_base.h"
#include "../facilities/waypoint_base.h"

#include "../table/strings.h"

#include "../vehicle/aircraft.h"
#include "../vehicle/roadveh.h"
#include "../vehicle/vehicle_base.h"
#include "../vehicle/vehicle_func.h"
#include "../vehicle/vehicle_gui.h"
#include "../vehicle/vehicle_logic.h"

#include "../cmd_helper.h"
#include "../command_func.h"
#include "../date_type.h"
#include "../news_func.h"

#include "../debug.h"
#include "../safeguards.h"





static CommandCost    DecloneOrder                    (Vehicle *dst, ProcessCmdFlag flags);





/* DestinationID must be at least as large as every these below, because it can
 * be any of them
 */
static_assert  (sizeof (DestinationID) >= sizeof (DepotID));
static_assert  (sizeof (DestinationID) >= sizeof (StationID));





/**
 * Add an order to the orderlist of a vehicle.
 * @param tile unused
 * @param flags operation to perform
 * @param p1 various bitstuffed elements
 * - p1 = (bit  0 - 19) - ID of the vehicle
 * - p1 = (bit 24 - 31) - the selected order (if any). If the last order is given,
 *                        the order will be inserted before that one
 *                        the maximum vehicle order id is 254.
 * @param p2 packed order to insert
 * @param text unused
 * @return the cost of this operation or an error
 */

CommandCost CmdInsertOrder (
	CommandContainer cmdvar
) {

	ProcessCmdFlag  flags       =                  cmdvar.dc_flags;
	VehicleOrderID  sel_ord     = (VehicleOrderID) cmdvar.var02;
	Order           new_order   = (Order)          cmdvar.var03;
	bool            first_order = false;



	Vehicle *vehicle = Vehicle::GetIfValid ((VehicleID)cmdvar.var01);

	if (vehicle == nullptr || !vehicle->IsPrimaryVehicle()) {

		return CMD_ERROR;
	}

	CommandCost ret = CheckOwnership (vehicle->owner);

	if (ret.Failed()) {

		return ret;
	}

	if (vehicle->order_list == nullptr) {

		first_order = true;
	}



	int switch_result = 0;

	/* Check if the inserted order is to the correct destination (owner, type),
	 * and has the correct flags if any */
	switch (new_order.GetType()) {

		case OT_GOTO_STATION: {

			switch_result = InsertStationOrder (vehicle, &new_order, &ret);

			break;
		}

		case OT_GOTO_DEPOT: {

			switch_result = InsertDepotOrder (vehicle, &new_order, &ret);

			break;
		}

		case OT_GOTO_WAYPOINT: {

			switch_result = InsertWaypointOrder (vehicle, &new_order, &ret);

			break;
		}

		case OT_CONDITIONAL: {

			switch_result = InsertConditionalOrder (vehicle, &new_order);

			break;
		}

		default: {

			return CMD_ERROR;
		}
	}



	if (switch_result == 1) {

		return CMD_ERROR;

	} else if (switch_result == 2) {

		return ret;

	} else if (switch_result == 3) {

		return_cmd_error (STR_ERROR_CAN_T_ADD_ORDER);

	} else if (switch_result == 4) {

		return_cmd_error (STR_ERROR_CAN_T_ADD_ORDER_SHARED);
	}



	if (sel_ord > GetNumOrders (vehicle)) {

		return CMD_ERROR;
	}

	if (GetNumOrders (vehicle) >= MAX_VEH_ORDER_ID) {

		return_cmd_error(STR_ERROR_TOO_MANY_ORDERS);
	}

	if (!Order::CanAllocateItem()) {

		return_cmd_error(STR_ERROR_NO_MORE_SPACE_FOR_ORDERS);
	}

	if (vehicle->order_list == nullptr && !OrderList::CanAllocateItem()) {

		return_cmd_error(STR_ERROR_NO_MORE_SPACE_FOR_ORDERS);
	}



	if (flags & DC_EXEC) {

		if (vehicle->vehicle_flags ^ VF_SHARED_ORDERS) {

			Order *new_o = new Order();

			new_o->AssignOrder (new_order);
			InsertOrder        (vehicle, new_o, sel_ord);

			if (first_order) {

				vehicle->current_order       = *new_o;
				vehicle->current_order_index = 0;

				UpdateVehicleDestination (

					vehicle,
					GetVehicleOrder (vehicle, vehicle->current_order_index)
				);
			}

		} else {

			Vehicle *next_vehicle = vehicle->FirstShared();
			DeleteOrderWarnings (next_vehicle);

			for (; next_vehicle != nullptr; next_vehicle = next_vehicle->NextShared()) {

				Order *new_o = new Order();

				new_o->AssignOrder (new_order);
				InsertOrder        (next_vehicle, new_o, sel_ord);

				if (first_order) {

					next_vehicle->current_order       = *new_o;
					next_vehicle->current_order_index = 0;

					UpdateVehicleDestination (

						next_vehicle,
						GetVehicleOrder (next_vehicle, next_vehicle->current_order_index)
					);
				}
			}
		}
	}

	return CommandCost();
}





int InsertStationOrder (
	Vehicle *vehicle, Order *new_order, CommandCost *ret
) {

	const Station *st = Station::GetIfValid (new_order->dest);

	if (st == nullptr) {

		return 1;
	}

	if (st->owner != OWNER_NONE) {

		*ret = CheckOwnership (st->owner);

		if (ret->Failed()) return 2;
	}

	if (!CanVehicleUseStation (vehicle, st)) {

		return 3;
	}

	for (Vehicle *u = vehicle->FirstShared(); u != nullptr; u = u->NextShared()) {

		if (!CanVehicleUseStation (u, st)) {

			return 4;
		}
	}

	/* Non stop only allowed for ground vehicles. */
	if (new_order->GetNonStopType() != ONSF_STOP_EVERYWHERE && !vehicle->IsGroundVehicle()) {

		return 1;
	}



	/* Filter invalid load/unload types. */
	switch (new_order->GetLoadType()) {

		case OLF_LOAD_IF_POSSIBLE: case OLFB_FULL_LOAD: case OLF_FULL_LOAD_ANY: case OLFB_NO_LOAD: {

			break;
		}

		default: {

			return 1;
		}
	}

		switch (new_order->GetUnloadType()) {

		case OUF_UNLOAD_IF_POSSIBLE: case OUFB_UNLOAD: case OUFB_TRANSFER: case OUFB_NO_UNLOAD: {

			break;
		}

		default: {

			return 1;
		}
	}



	/* Filter invalid stop locations */
	switch (new_order->GetStopLocation()) {

		case OSL_PLATFORM_NEAR_END:
		case OSL_PLATFORM_MIDDLE: {

			if (vehicle->type != VEH_TRAIN) {

				return 1;
			}

			FALLTHROUGH;
		}

		case OSL_PLATFORM_FAR_END: {

			break;
		}

		default: {

			return 1;
		}
	}

	return 0;
}






int InsertDepotOrder (
	Vehicle *vehicle, Order *new_order, CommandCost *ret
) {

	if ((new_order->GetDepotActionType() & ODATFB_NEAREST_DEPOT) == 0) {

		if (vehicle->type == VEH_AIRCRAFT) {

			const Station *st = Station::GetIfValid (new_order->dest);

			if (st == nullptr) {

				return 1;
			}

			CommandCost ret = CheckOwnership (st->owner);

			if (ret.Failed()) {

				return 2;
			}

			if (!CanVehicleUseStation (vehicle, st) || !st->airport.HasHangar()) {

				return 1;
			}

		} else {

			const Depot *dp = Depot::GetIfValid (new_order->dest);

			if (dp == nullptr) {

				return 1;
			}

			CommandCost ret = CheckOwnership (GetTileOwner(dp->xy));

			if (ret.Failed()) {

				return 2;
			}



			switch (vehicle->type) {

				case VEH_TRAIN: {

					if (!IsRailDepotTile (dp->xy)) {

						return 1;
					}

					break;
				}

				case VEH_ROAD: {

					if (!IsRoadDepotTile (dp->xy)) {

						return 1;
					}

					break;
				}

				case VEH_SHIP: {

					if (!IsShipDepotTile (dp->xy)) {

						return 1;
					}

					break;
				}

				default: {

					return 1;
				}
			}


		}
	}



	if (new_order->GetNonStopType() != ONSF_STOP_EVERYWHERE && !vehicle->IsGroundVehicle()) {

		return 1;
	}

	if (new_order->GetDepotOrderType()
		& ~(ODTFB_PART_OF_ORDERS
			| ((new_order->GetDepotOrderType() & ODTFB_PART_OF_ORDERS) != 0
			? ODTFB_SERVICE : 0)
		   )
	) {

		return 1;
	}

	if (new_order->GetDepotActionType() & ~(ODATFB_HALT | ODATFB_NEAREST_DEPOT)) {

		return 1;
	}

	if (   (new_order->GetDepotOrderType()  & ODTFB_SERVICE)
		&& (new_order->GetDepotActionType() & ODATFB_HALT)
	) {

		return 1;
	}

	return 0;
}





int InsertWaypointOrder (
	Vehicle *vehicle, Order *new_order, CommandCost *ret
) {

	const Waypoint *wp = Waypoint::GetIfValid(new_order->dest);

	if (wp == nullptr) {

		return 1;
	}



	switch (vehicle->type) {

		case VEH_TRAIN: {

			if (!(wp->facilities & FACIL_TRAIN)) {

				return 3;
			}

			CommandCost ret = CheckOwnership(wp->owner);

			if (ret.Failed()) {

				return 2;
			}

			break;
		}

		case VEH_SHIP: {

			if (!(wp->facilities & FACIL_DOCK)) {

				return 3;
			}

			if (wp->owner != OWNER_NONE) {

				CommandCost ret = CheckOwnership(wp->owner);

				if (ret.Failed()) {

					return 2;
				}
			}

			break;
		}

		default: {

			return 1;
		}
	}



	/* Order flags can be any of the following for waypoints:
	 * [non-stop]
	 * non-stop orders (if any) are only valid for trains */
	if (new_order->GetNonStopType() != ONSF_STOP_EVERYWHERE
		&& vehicle->type != VEH_TRAIN
	) {

		return 1;
	}

	return 0;
}





int InsertConditionalOrder (
	Vehicle *vehicle, Order *new_order
) {

	VehicleOrderID skip_to = new_order->GetConditionSkipToOrder();

	// Always allow jumping to the first even if there is no order
	if (skip_to != 0 && skip_to >= GetNumOrders (vehicle)) {

		return 1;
	}

	if (new_order->GetConditionVariable() >= OCV_END) {

		return 1;
	}

	OrderConditionComparator occ = new_order->GetConditionComparator();

	if (occ >= OCC_END) {

		return 1;
	}



	switch (new_order->GetConditionVariable()) {

		case OCV_REQUIRES_SERVICE: {

			if (occ != OCC_IS_TRUE && occ != OCC_IS_FALSE) {

				return 1;
			}

			break;
		}

		case OCV_UNCONDITIONALLY: {

			if (occ != OCC_EQUALS) {

				return 1;
			}

			if (new_order->GetConditionValue() != 0) {

				return 1;
			}

			break;
		}

		case OCV_LOAD_PERCENTAGE:
		case OCV_RELIABILITY:     {

			if (new_order->GetConditionValue() > 100) {

				return 1;
			}

			FALLTHROUGH;
		}

		default: {

			if (occ == OCC_IS_TRUE || occ == OCC_IS_FALSE) {

				return 1;
			}

			break;
		}
	}



	return 0;
}





/**
 * Delete an order from the orderlist of a vehicle.
 * @param tile unused
 * @param flags operation to perform
 * @param p1 the ID of the vehicle
 * @param p2 the order to delete (max 255)
 * @param text unused
 * @return the cost of this operation or an error
 */
CommandCost CmdDeleteOrder(CommandContainer cmdvar)
{
	ProcessCmdFlag flags = cmdvar.dc_flags;
	uint32 p1 = cmdvar.p1;
	uint32 p2 = cmdvar.p2;
	const std::string text = cmdvar.string;

	VehicleID veh_id = GB(p1, 0, 20);
	VehicleOrderID sel_ord = GB(p2, 0, 8);

	Vehicle *v = Vehicle::GetIfValid(veh_id);

	if (v == nullptr || !v->IsPrimaryVehicle()) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	/* If we did not select an order, we maybe want to de-clone the orders */
	if (sel_ord >= GetNumOrders (v)) return DecloneOrder(v, flags);

	if (GetVehicleOrder (v, sel_ord) == nullptr) return CMD_ERROR;

	if (flags & DC_EXEC) {
		if (v->vehicle_flags ^ VF_SHARED_ORDERS) {
			DeleteOrder(v, sel_ord);
		} else {
			for (Vehicle *u = v->FirstShared(); u != nullptr; u = u->NextShared()) {
				DeleteOrder(u, sel_ord);
			}
		}
	}
	return CommandCost();
}





/**
 * Goto order of order-list.
 * @return the cost of this operation or an error
 */

CommandCost CmdSkipToOrder (
	CommandContainer cmdvar
) {

	ProcessCmdFlag     flags           = cmdvar.dc_flags;
	Vehicle            *vehicle        = Vehicle::GetIfValid (cmdvar.var01);
	uint32             selected_order  = cmdvar.var02;
	bool               ctrl_pressed    = cmdvar.var03;
	const std::string  text            = cmdvar.string;

	if (
		vehicle == nullptr
		|| !vehicle->IsPrimaryVehicle()
		|| (int32)selected_order - 1 > GetNumOrders (vehicle)
		|| GetNumOrders (vehicle) < 2
	) {

		return CMD_ERROR;
	}

	CommandCost ret = CheckOwnership (vehicle->owner);

	if (ret.Failed()) {

		return ret;
	}

	if (flags & DC_EXEC) {

		vehicle->vehstatus &= ~VS_WAITING;

		if (vehicle->vehicle_location & VL_AT_STATION) {

			vehicle->vehicle_location |= VL_LEAVE_STATION;

		} else if (vehicle->vehicle_location & VL_AT_WAYPOINT) {

			vehicle->vehicle_location |= VL_LEAVE_WAYPOINT;

		} else if (vehicle->vehicle_location & VL_AT_DEPOT) {

			vehicle->vehicle_location |= VL_LEAVE_DEPOT;
		}

		if (ctrl_pressed) {

			vehicle->current_order_index = selected_order;

		} else {

			if ((vehicle->current_order_index + 1) > vehicle->order_list->num_orders -1) {

				vehicle->current_order_index = 0;

			} else {

				++vehicle->current_order_index;
			}
		}

		InvalidateVehicleOrder (vehicle, VIWD_MODIFY_ORDERS);

		/* If we have an aircraft/ship, they have a mini-schedule so update them all */
		if (vehicle->type == VEH_TRAIN)    ((Train *)vehicle)->force_proceed = TFP_NONE;
		if (vehicle->type == VEH_AIRCRAFT) SetWindowClassesDirty (WC_AIRCRAFT_LIST);
		if (vehicle->type == VEH_SHIP)     SetWindowClassesDirty (WC_SHIPS_LIST);
	}

	return CommandCost();
}





/**
 * Move an order inside the orderlist
 * @param tile unused
 * @param flags operation to perform
 * @param p1 the ID of the vehicle
 * @param p2 order to move and target
 *           bit 0-15  : the order to move
 *           bit 16-31 : the target order
 * @param text unused
 * @return the cost of this operation or an error
 * @note The target order will move one place down in the orderlist
 *  if you move the order upwards else it'll move it one place down
 */
CommandCost CmdMoveOrder(CommandContainer cmdvar)
{
	ProcessCmdFlag flags = cmdvar.dc_flags;
	uint32 p1 = cmdvar.p1;
	uint32 p2 = cmdvar.p2;
	const std::string text = cmdvar.string;

	VehicleID veh = GB(p1, 0, 20);
	VehicleOrderID moving_order = GB(p2,  0, 16);
	VehicleOrderID target_order = GB(p2, 16, 16);

	Vehicle *v = Vehicle::GetIfValid(veh);
	if (v == nullptr || !v->IsPrimaryVehicle()) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	/* Don't make senseless movements */
	if (moving_order >= GetNumOrders (v) || target_order >= GetNumOrders (v) ||
			moving_order == target_order || GetNumOrders (v) <= 1) return CMD_ERROR;

	Order *moving_one = GetVehicleOrder (v, moving_order);
	/* Don't move an empty order */
	if (moving_one == nullptr) return CMD_ERROR;

	if (flags & DC_EXEC) {
		if (v->vehicle_flags ^ VF_SHARED_ORDERS) {
			MoveOrder (v->order_list, moving_order, target_order);
		} else {
			for (Vehicle *u = v->FirstShared(); u != nullptr; u = u->NextShared()) {
				MoveOrder (u->order_list, moving_order, target_order);
			}
		}

		/* Update shared list */
		Vehicle *u = v->FirstShared();

		DeleteOrderWarnings(u);

		for (; u != nullptr; u = u->NextShared()) {
			/* Update the current order.
			 * There are multiple ways to move orders, which result in cur_implicit_order_index
			 * and current_order_index to not longer make any sense. E.g. moving another
			 * real order between them.
			 *
			 * Basically one could choose to preserve either of them, but not both.
			 * While both ways are suitable in this or that case from a human point of view, neither
			 * of them makes really sense.
			 * However, from an AI point of view, preserving current_order_index is the most
			 * predictable and transparent behaviour.
			 *
			 * With that decision it basically does not matter what we do to cur_implicit_order_index.
			 * If we change orders between the implicit- and real-index, the implicit orders are mostly likely
			 * completely out-dated anyway. So, keep it simple and just keep cur_implicit_order_index as well.
			 * The worst which can happen is that a lot of implicit orders are removed when reaching current_order.
			 */
			if (u->current_order_index == moving_order) {
				u->current_order_index = target_order;
			} else if (u->current_order_index > moving_order && u->current_order_index <= target_order) {
				u->current_order_index--;
			} else if (u->current_order_index < moving_order && u->current_order_index >= target_order) {
				u->current_order_index++;
			}

			assert(v->order_list == u->order_list);
			/* Update any possible open window of the vehicle */
			InvalidateVehicleOrder(u, moving_order | (target_order << 8));
		}

		/* As we move an order, the order to skip to will be 'wrong'. */
		for (Order *order : VehicleOrders (v)) {
			if (order->IsType(OT_CONDITIONAL)) {
				VehicleOrderID order_id = order->GetConditionSkipToOrder();
				if (order_id == moving_order) {
					order_id = target_order;
				} else if (order_id > moving_order && order_id <= target_order) {
					order_id--;
				} else if (order_id < moving_order && order_id >= target_order) {
					order_id++;
				}
				order->SetConditionSkipToOrder(order_id);
			}
		}

		/* Make sure to rebuild the whole list */
		InvalidateWindowClassesData(GetWindowClassForVehicleType(v->type), 0);
	}

	return CommandCost();
}





/**
 * Modify an order in the orderlist of a vehicle.
 * @param tile unused
 * @param flags operation to perform
 * @param p1 various bitstuffed elements
 * - p1 = (bit  0 - 19) - ID of the vehicle
 * - p1 = (bit 24 - 31) - the selected order (if any). If the last order is given,
 *                        the order will be inserted before that one
 *                        the maximum vehicle order id is 254.
 * @param p2 various bitstuffed elements
 *  - p2 = (bit 0 -  3) - what data to modify (@see ModifyOrderFlags)
 *  - p2 = (bit 4 - 15) - the data to modify
 * @param text unused
 * @return the cost of this operation or an error
 */
CommandCost CmdModifyOrder(CommandContainer cmdvar)
{
	ProcessCmdFlag flags = cmdvar.dc_flags;
	uint32 p1 = cmdvar.p1;
	uint32 p2 = cmdvar.p2;
	const std::string text = cmdvar.string;

	VehicleOrderID sel_ord = GB(p1, 20,  8);
	VehicleID veh          = GB(p1,  0, 20);
	ModifyOrderFlags mof   = Extract<ModifyOrderFlags, 0, 4>(p2);
	uint16 data            = GB(p2,  4, 11);

	if (mof >= MOF_END) return CMD_ERROR;

	Vehicle *v = Vehicle::GetIfValid(veh);
	if (v == nullptr || !v->IsPrimaryVehicle()) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	/* Is it a valid order? */
	if (sel_ord >= GetNumOrders (v)) return CMD_ERROR;

	for (Vehicle *u = v->FirstShared(); u != nullptr; u = u->NextShared()) { // begin main for loop

	Order *order = GetVehicleOrder (u, sel_ord);
	switch (order->GetType()) {
		case OT_GOTO_STATION:
			if (mof != MOF_NON_STOP && mof != MOF_STOP_LOCATION && mof != MOF_UNLOAD && mof != MOF_LOAD) return CMD_ERROR;
			break;

		case OT_GOTO_DEPOT:
			if (mof != MOF_NON_STOP && mof != MOF_DEPOT_ACTION) return CMD_ERROR;
			break;

		case OT_GOTO_WAYPOINT:
			if (mof != MOF_NON_STOP && mof != MOF_WAYPOINT_FLAGS) return CMD_ERROR;
			break;

		case OT_CONDITIONAL:
			if (mof != MOF_COND_VARIABLE && mof != MOF_COND_COMPARATOR && mof != MOF_COND_VALUE && mof != MOF_COND_DESTINATION) return CMD_ERROR;
			break;

		default:
			return CMD_ERROR;
	}

	switch (mof) {
		default: NOT_REACHED();

		case MOF_NON_STOP:
			if (!u->IsGroundVehicle()) return CMD_ERROR;
			if (data >= ONSF_END) return CMD_ERROR;
			if (data == order->GetNonStopType()) return CMD_ERROR;
			break;

		case MOF_STOP_LOCATION:
			if (u->type != VEH_TRAIN) return CMD_ERROR;
			if (data >= OSL_END) return CMD_ERROR;
			break;

		case MOF_UNLOAD:
			if (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION) return CMD_ERROR;
			if ((data & ~(OUFB_UNLOAD | OUFB_TRANSFER | OUFB_NO_UNLOAD)) != 0) return CMD_ERROR;
			/* Unload and no-unload are mutual exclusive and so are transfer and no unload. */
			if (data != 0 && ((data & (OUFB_UNLOAD | OUFB_TRANSFER)) != 0) == ((data & OUFB_NO_UNLOAD) != 0)) return CMD_ERROR;
			if (data == order->GetUnloadType()) return CMD_ERROR;
			break;

		case MOF_LOAD:
			if (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION) return CMD_ERROR;
			if (data > OLFB_NO_LOAD || data == 1) return CMD_ERROR;
			if (data == order->GetLoadType()) return CMD_ERROR;
			break;

		case MOF_DEPOT_ACTION:
			if (data >= DA_END) return CMD_ERROR;
			break;

		case MOF_COND_VARIABLE:
			if (data >= OCV_END) return CMD_ERROR;
			break;

		case MOF_COND_COMPARATOR:
			if (data >= OCC_END) return CMD_ERROR;
			switch (order->GetConditionVariable()) {
				case OCV_UNCONDITIONALLY: return CMD_ERROR;

				case OCV_REQUIRES_SERVICE:
					if (data != OCC_IS_TRUE && data != OCC_IS_FALSE) return CMD_ERROR;
					break;

				default:
					if (data == OCC_IS_TRUE || data == OCC_IS_FALSE) return CMD_ERROR;
					break;
			}
			break;

		case MOF_COND_VALUE:
			switch (order->GetConditionVariable()) {
				case OCV_UNCONDITIONALLY:
				case OCV_REQUIRES_SERVICE:
					return CMD_ERROR;

				case OCV_LOAD_PERCENTAGE:
				case OCV_RELIABILITY:
					if (data > 100) return CMD_ERROR;
					break;

				default:
					if (data > 2047) return CMD_ERROR;
					break;
			}
			break;

		case MOF_COND_DESTINATION:
			if (data >= GetNumOrders (u)) return CMD_ERROR;
			break;

		case MOF_WAYPOINT_FLAGS:
			order->SetWaypointFlags((OrderWaypointFlags)data);
			break;
	}

	if (flags & DC_EXEC) {
		switch (mof) {
			case MOF_NON_STOP:
				order->SetNonStopType((OrderNonStopFlags)data);
				if (data & ONSF_NO_STOP_AT_DESTINATION_STATION) {
					order->SetRefit(CT_NO_REFIT);
					order->SetLoadType(OLFB_NO_LOAD);
					order->SetUnloadType(OUFB_NO_UNLOAD);
				}
				break;

			case MOF_STOP_LOCATION:
				order->SetStopLocation((OrderStopLocation)data);
				break;

			case MOF_UNLOAD:
				order->SetUnloadType((OrderUnloadFlags)data);
				break;

			case MOF_LOAD:
				order->SetLoadType((OrderLoadFlags)data);
				if (data & OLFB_NO_LOAD) order->SetRefit(CT_NO_REFIT);
				break;

			case MOF_DEPOT_ACTION: {
				switch (data) {
					case DA_ALWAYS_GO:
						order->SetDepotOrderType((OrderDepotTypeFlags)(order->GetDepotOrderType() & ~ODTFB_SERVICE));
						order->SetDepotActionType((OrderDepotActionFlags)(order->GetDepotActionType() & ~ODATFB_HALT));
						break;

					case DA_SERVICE:
						order->SetDepotOrderType((OrderDepotTypeFlags)(order->GetDepotOrderType() | ODTFB_SERVICE));
						order->SetDepotActionType((OrderDepotActionFlags)(order->GetDepotActionType() & ~ODATFB_HALT));
						order->SetRefit(CT_NO_REFIT);
						break;

					case DA_STOP:
						order->SetDepotOrderType((OrderDepotTypeFlags)(order->GetDepotOrderType() & ~ODTFB_SERVICE));
						order->SetDepotActionType((OrderDepotActionFlags)(order->GetDepotActionType() | ODATFB_HALT));
						order->SetRefit(CT_NO_REFIT);
						break;

					case DA_PARK:
						order->SetDepotOrderType((OrderDepotTypeFlags)(order->GetDepotOrderType() & ~ODTFB_SERVICE));
						order->SetDepotActionType((OrderDepotActionFlags)(order->GetDepotActionType() | ODATFB_PARK));
						order->SetRefit(CT_NO_REFIT);
						break;

					default:
						NOT_REACHED();
				}
				break;
			}

			case MOF_COND_VARIABLE: {
				order->SetConditionVariable((OrderConditionVariable)data);

				OrderConditionComparator occ = order->GetConditionComparator();
				switch (order->GetConditionVariable()) {
					case OCV_UNCONDITIONALLY:
						order->SetConditionComparator(OCC_EQUALS);
						order->SetConditionValue(0);
						break;

					case OCV_REQUIRES_SERVICE:
						if (occ != OCC_IS_TRUE && occ != OCC_IS_FALSE) order->SetConditionComparator(OCC_IS_TRUE);
						order->SetConditionValue(0);
						break;

					case OCV_LOAD_PERCENTAGE:
					case OCV_RELIABILITY:
						if (order->GetConditionValue() > 100) order->SetConditionValue(100);
						FALLTHROUGH;

					default:
						if (occ == OCC_IS_TRUE || occ == OCC_IS_FALSE) order->SetConditionComparator(OCC_EQUALS);
						break;
				}
				break;
			}

			case MOF_COND_COMPARATOR:
				order->SetConditionComparator((OrderConditionComparator)data);
				break;

			case MOF_COND_VALUE:
				order->SetConditionValue(data);
				break;

			case MOF_COND_DESTINATION:
				order->SetConditionSkipToOrder(data);
				break;

			case MOF_WAYPOINT_FLAGS:
				order->SetWaypointFlags((OrderWaypointFlags)data);
				break;

			default: NOT_REACHED();
		}

		/* Update the windows and full load flags, also for vehicles that share the same order list */
			DeleteOrderWarnings(u);
			/* Toggle u->current_order "Full load" flag if it changed.
			 * However, as the same flag is used for depot orders, check
			 * whether we are not going to a depot as there are three
			 * cases where the full load flag can be active and only
			 * one case where the flag is used for depot orders. In the
			 * other cases for the OrderType the flags are not used,
			 * so do not care and those orders should not be active
			 * when this function is called.
			 */
			if (sel_ord == u->current_order_index &&
					(u->current_order.IsType(OT_GOTO_STATION) || u->current_order.IsType(OT_LOADING)) &&
					u->current_order.GetLoadType() != order->GetLoadType()) {
				u->current_order.SetLoadType(order->GetLoadType());
			}
			InvalidateVehicleOrder(u, VIWD_MODIFY_ORDERS);
	}

	} // end main for loop

	return CommandCost();
}





/**
 * Clone/share/copy an order-list of another vehicle.
 * @param tile unused
 * @param flags operation to perform
 * @param p1 various bitstuffed elements
 * - p1 = (bit  0-19) - destination vehicle to clone orders to
 * - p1 = (bit 30-31) - action to perform
 * @param p2 source vehicle to clone orders from, if any (none for CO_UNSHARE)
 * @param text unused
 * @return the cost of this operation or an error
 */
CommandCost CmdCloneOrder(CommandContainer cmdvar)
{
	ProcessCmdFlag flags = cmdvar.dc_flags;
	uint32 p1 = cmdvar.p1;
	uint32 p2 = cmdvar.p2;
	const std::string text = cmdvar.string;

	VehicleID veh_src = GB(p2, 0, 20);
	VehicleID veh_dst = GB(p1, 0, 20);

	Vehicle *dst = Vehicle::GetIfValid(veh_dst);
	if (dst == nullptr || !dst->IsPrimaryVehicle()) return CMD_ERROR;

	CommandCost ret = CheckOwnership(dst->owner);
	if (ret.Failed()) return ret;

	switch (GB(p1, 30, 2)) {
		case CO_SHARE: {
			Vehicle *src = Vehicle::GetIfValid(veh_src);

			/* Sanity checks */
			if (src == nullptr || !src->IsPrimaryVehicle() || dst->type != src->type || dst == src) return CMD_ERROR;

			CommandCost ret = CheckOwnership(src->owner);
			if (ret.Failed()) return ret;

			/* Trucks can't share orders with busses (and visa versa) */
			if (src->type == VEH_ROAD && RoadVehicle::From(src)->IsBus() != RoadVehicle::From(dst)->IsBus()) {
				return CMD_ERROR;
			}

			/* Is the vehicle already in the shared list? */
			if (src->FirstShared() == dst->FirstShared()) return CMD_ERROR;

			for (const Order *order : VehicleOrders (src)) {
				if (!OrderGoesToStation(dst, order)) continue;

				/* Allow copying unreachable destinations if they were already unreachable for the source.
				 * This is basically to allow cloning / autorenewing / autoreplacing vehicles, while the stations
				 * are temporarily invalid due to reconstruction. */
				const Station *st = Station::Get(order->dest);
				if (CanVehicleUseStation(src, st) && !CanVehicleUseStation(dst, st)) {
					return_cmd_error(STR_ERROR_CAN_T_COPY_SHARE_ORDER);
				}
			}

			/* Check for aircraft range limits. */
			if (dst->type == VEH_AIRCRAFT && !CheckAircraftOrderDistance(Aircraft::From(dst), src, src->order_list->first)) {
				return_cmd_error(STR_ERROR_AIRCRAFT_NOT_ENOUGH_RANGE);
			}

			if (src->order_list == nullptr && !OrderList::CanAllocateItem()) {
				return_cmd_error(STR_ERROR_NO_MORE_SPACE_FOR_ORDERS);
			}

			if (flags & DC_EXEC) {
				Order *first = nullptr;
				Order **order_dst;

				/* If the destination vehicle had a OrderList, destroy it.
				 * We only reset the order indices, if the new orders are obviously different.
				 * (We mainly do this to keep the order indices valid and in range.) */
				DeleteVehicleOrders(dst, true, GetNumOrders (dst) != GetNumOrders (src));

				order_dst = &first;
				for (const Order *order : VehicleOrders (src)) {
					*order_dst = new Order();
					(*order_dst)->AssignOrder(*order);
					order_dst = &(*order_dst)->next;
				}
				if (dst->order_list == nullptr) {

					dst->order_list = new OrderList;
					CreateOrderList (dst->order_list, first, dst);

				} else {

					assert(dst->order_list->first == nullptr);
					assert(!dst->order_list->num_vehicles > 1);
					delete dst->order_list;
					assert(OrderList::CanAllocateItem());
					dst->order_list = new OrderList;
					CreateOrderList (dst->order_list, first, dst);
				}

				/* Link this vehicle in the shared-list */
				AddToShared(src, dst);

				InvalidateVehicleOrder(dst, VIWD_REMOVE_ALL_ORDERS);

				InvalidateWindowClassesData(GetWindowClassForVehicleType(dst->type), 0);
			}
			break;
		}

		case CO_COPY: {
			Vehicle *src = Vehicle::GetIfValid(veh_src);

			/* Sanity checks */
			if (src == nullptr || !src->IsPrimaryVehicle() || dst->type != src->type || dst == src) return CMD_ERROR;

			CommandCost ret = CheckOwnership(src->owner);
			if (ret.Failed()) return ret;

			/* Trucks can't copy all the orders from busses (and visa versa),
			 * and neither can helicopters and aircraft. */
			for (const Order *order : VehicleOrders (src)) {
				if (OrderGoesToStation(dst, order) &&
						!CanVehicleUseStation(dst, Station::Get(order->dest))) {
					return_cmd_error(STR_ERROR_CAN_T_COPY_SHARE_ORDER);
				}
			}

			/* Check for aircraft range limits. */
			if (dst->type == VEH_AIRCRAFT && !CheckAircraftOrderDistance(Aircraft::From(dst), src, src->order_list->first)) {
				return_cmd_error(STR_ERROR_AIRCRAFT_NOT_ENOUGH_RANGE);
			}

			/* make sure there are orders available */
			if (!Order::CanAllocateItem(GetNumOrders (src)) || !OrderList::CanAllocateItem()) {
				return_cmd_error(STR_ERROR_NO_MORE_SPACE_FOR_ORDERS);
			}

			if (flags & DC_EXEC) {
				Order *first = nullptr;
				Order **order_dst;

				/* If the destination vehicle had an order list, destroy the chain but keep the OrderList.
				 * We only reset the order indices, if the new orders are obviously different.
				 * (We mainly do this to keep the order indices valid and in range.) */
				DeleteVehicleOrders(dst, true, GetNumOrders (dst) != GetNumOrders (src));

				order_dst = &first;
				for (const Order *order : VehicleOrders (src)) {
					*order_dst = new Order();
					(*order_dst)->AssignOrder(*order);
					order_dst = &(*order_dst)->next;
				}

				if (dst->order_list == nullptr) {

					dst->order_list = new OrderList;
					CreateOrderList (dst->order_list, first, dst);

				} else {

					assert(dst->order_list->first == nullptr);
					assert(!dst->order_list->num_vehicles > 1);
					delete dst->order_list;
					assert(OrderList::CanAllocateItem());
					dst->order_list = new OrderList;
					CreateOrderList (dst->order_list, first, dst);
				}

				InvalidateVehicleOrder(dst, VIWD_REMOVE_ALL_ORDERS);

				InvalidateWindowClassesData(GetWindowClassForVehicleType(dst->type), 0);
			}
			break;
		}

		case CO_UNSHARE: return DecloneOrder(dst, flags);
		default: return CMD_ERROR;
	}

	return CommandCost();
}





/**
 * Declone an order-list
 * @param *dst delete the orders of this vehicle
 * @param flags execution flags
 */
static CommandCost DecloneOrder(Vehicle *v, ProcessCmdFlag flags)
{
	if (flags & DC_EXEC) {
		RemoveFromShared(v);

		InvalidateVehicleOrder(v, VIWD_REMOVE_ALL_ORDERS);
		InvalidateWindowClassesData(GetWindowClassForVehicleType(v->type), 0);
	}

	return CommandCost();
}





/**
 * Add/remove refit orders from an order
 * @param tile Not used
 * @param flags operation to perform
 * @param p1 VehicleIndex of the vehicle having the order
 * @param p2 bitmask
 *   - bit 0-7 CargoID
 *   - bit 16-23 number of order to modify
 * @param text unused
 * @return the cost of this operation or an error
 */
CommandCost CmdOrderRefit(CommandContainer cmdvar)
{
	ProcessCmdFlag flags = cmdvar.dc_flags;
	uint32 p1 = cmdvar.p1;
	uint32 p2 = cmdvar.p2;
	const std::string text = cmdvar.string;

	VehicleID veh = GB(p1, 0, 20);
	VehicleOrderID order_number  = GB(p2, 16, 8);
	CargoID cargo = GB(p2, 0, 8);

	if (cargo >= NUM_CARGO && cargo != CT_NO_REFIT && cargo != CT_AUTO_REFIT) return CMD_ERROR;

	const Vehicle *v = Vehicle::GetIfValid(veh);
	if (v == nullptr || !v->IsPrimaryVehicle()) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	Order *order = GetVehicleOrder (v, order_number);
	if (order == nullptr) return CMD_ERROR;

	/* Automatic refit cargo is only supported for goto station orders. */
	if (cargo == CT_AUTO_REFIT && !order->IsType(OT_GOTO_STATION)) return CMD_ERROR;

	if (order->GetLoadType() & OLFB_NO_LOAD) return CMD_ERROR;

	if (flags & DC_EXEC) {
		order->SetRefit(cargo);

		/* Make the depot order an 'always go' order. */
		if (cargo != CT_NO_REFIT && order->IsType(OT_GOTO_DEPOT)) {
			order->SetDepotOrderType((OrderDepotTypeFlags)(order->GetDepotOrderType() & ~ODTFB_SERVICE));
			order->SetDepotActionType((OrderDepotActionFlags)(order->GetDepotActionType() & ~ODATFB_HALT));
		}

		for (Vehicle *u = v->FirstShared(); u != nullptr; u = u->NextShared()) {
			/* Update any possible open window of the vehicle */
			InvalidateVehicleOrder(u, VIWD_MODIFY_ORDERS);

			/* If the vehicle already got the current depot set as current order, then update current order as well */
			if (u->current_order_index == order_number && (u->current_order.GetDepotOrderType() & ODTFB_PART_OF_ORDERS)) {
				u->current_order.SetRefit(cargo);
			}
		}
	}

	return CommandCost();
}
