/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file order.cpp Handling of order functions. */





#include "../stdafx.h"

#include "order_logic.h"
#include "order.h"
#include "order_backup.h"

#include "vehicle/engine_base.h"
#include "vehicle/vehicle_base.h"
#include "vehicle/vehicle_func.h"
#include "vehicle/vehicle_gui.h"





/**
 * Checks whether the order goes to a station or not, i.e. whether the
 * destination is a station
 * @param v the vehicle to check for
 * @param o the order to check
 * @return true if the destination is a station
 */

bool OrderGoesToStation (
	const Vehicle *v, const Order *o
) {

	return o->IsType(OT_GOTO_STATION) ||
			(v->type == VEH_AIRCRAFT && o->IsType(OT_GOTO_DEPOT) && !(o->GetDepotActionType() & ODATFB_NEAREST_DEPOT));
}





/**
 * Process a conditional order and determine the next order.
 * @param order the order the vehicle currently has
 * @param v the vehicle to update
 * @return index of next order to jump to, or INVALID_VEH_ORDER_ID to use the next order
 */

VehicleOrderID ProcessConditionalOrder (
	const Order *order, const Vehicle *v
) {

	if (order->GetType() != OT_CONDITIONAL) {

		return INVALID_VEH_ORDER_ID;
	}

	bool                      skip_order  = false;
	OrderConditionComparator  occ         = order->GetConditionComparator();
	uint16                    value       = order->GetConditionValue();

	switch (order->GetConditionVariable()) {

		case OCV_UNCONDITIONALLY: {

			skip_order = true;

			break;
		}

		case OCV_LOAD_PERCENTAGE: {

			skip_order = OrderConditionCompare (occ, CalcPercentVehicleFilled (v, nullptr), value);

			break;
		}

		case OCV_RELIABILITY: {

			skip_order = OrderConditionCompare (occ, ToPercent16 (v->reliability), value);

			break;
		}

		case OCV_MAX_RELIABILITY: {

			skip_order = OrderConditionCompare (occ,
												ToPercent16 (Engine::Get (v->engine_type)->reliability),
												value
											   )
			;

			break;
		}

		case OCV_MAX_SPEED: {

			skip_order = OrderConditionCompare (occ, v->GetDisplayMaxSpeed() * 1000 / 1609, value);

			break;
		}

		case OCV_AGE: {

			skip_order = OrderConditionCompare (occ, v->age / DAYS_IN_LEAP_YEAR, value);

			break;
		}

		case OCV_REQUIRES_SERVICE: {

			skip_order = OrderConditionCompare (occ, v->NeedsServicing(), value);

			break;
		}

		case OCV_REMAINING_LIFETIME: {

			skip_order = OrderConditionCompare (occ,
												std::max (v->max_age - v->age + DAYS_IN_LEAP_YEAR - 1, 0)
													/ DAYS_IN_LEAP_YEAR,
												value
											   )
			;

			break;
		}

		default: {

			NOT_REACHED();
		}
	}

	return skip_order ? order->GetConditionSkipToOrder() : (VehicleOrderID)INVALID_VEH_ORDER_ID;
}





/**
 * Compare the variable and value based on the given comparator.
 */

bool OrderConditionCompare (
	OrderConditionComparator occ, int variable, int value
) {

	switch (occ) {
		case OCC_EQUALS:      return variable == value;
		case OCC_NOT_EQUALS:  return variable != value;
		case OCC_LESS_THAN:   return variable <  value;
		case OCC_LESS_EQUALS: return variable <= value;
		case OCC_MORE_THAN:   return variable >  value;
		case OCC_MORE_EQUALS: return variable >= value;
		case OCC_IS_TRUE:     return variable != 0;
		case OCC_IS_FALSE:    return variable == 0;

		default: NOT_REACHED();
	}
}





/**
 * Removes an order from all vehicles. Triggers when, say, a station is removed.
 * @param type The type of the order (OT_GOTO_[STATION|DEPOT|WAYPOINT]).
 * @param destination The destination. Can be a StationID, DepotID or WaypointID.
 * @param hangar Only used for airports in the destination.
 *               When false, remove airport and hangar orders.
 *               When true, remove either airport or hangar order.
 */

void RemoveOrderFromAllVehicles (
	OrderType type, DestinationID destination, bool hangar
) {

	/* Aircraft have StationIDs for depot orders and never use DepotIDs
	 * This fact is handled specially below
	 */

	/* Go through all vehicles */
	for (Vehicle *v : Vehicle::Iterate()) {
		Order *order;

		order = &v->current_order;
//		if ((v->type == VEH_AIRCRAFT && order->IsType(OT_GOTO_DEPOT) && !hangar ? OT_GOTO_STATION : order->GetType()) == type &&
//				(!hangar || v->type == VEH_AIRCRAFT) && v->current_order.dest == destination) {
//			order->MakeDummy();
//			SetWindowDirty(WC_VEHICLE_VIEW, v->index);
//		}

		/* Clear the order from the order-list */
		int id = -1;
		for (Order *order : VehicleOrders (v)) {
			id++;

			OrderType ot = order->GetType();
			if (ot == OT_GOTO_DEPOT && (order->GetDepotActionType() & ODATFB_NEAREST_DEPOT) != 0) continue;
			if (ot == OT_GOTO_DEPOT && hangar && v->type != VEH_AIRCRAFT) continue; // Not an aircraft? Can't have a hangar order.
			if ((v->type == VEH_AIRCRAFT && ot == OT_GOTO_DEPOT && !hangar)) ot = OT_GOTO_STATION;
			if (ot == type && order->dest == destination) {

				/* Clear wait time */
//				v->order_list->UpdateTimetableDuration(-order->wait_time);
//				if (order->IsWaitTimetabled()) {
//					v->order_list->UpdateTimetableDuration(-order->wait_time);
//					order->wait_time = 0;
//				}
				order->wait_time = 0;

				/* Clear order, preserving travel time */
//				bool travel_timetabled = order->IsTravelTimetabled();
//				order->MakeDummy();
//				order->travel_time = travel_timetabled;

				for (const Vehicle *w = v->FirstShared(); w != nullptr; w = w->NextShared()) {
					/* In GUI, simulate by removing the order and adding it back */
					InvalidateVehicleOrder(w, id | (INVALID_VEH_ORDER_ID << 8));
					InvalidateVehicleOrder(w, (INVALID_VEH_ORDER_ID << 8) | id);
				}
			}
		}
	}

	OrderBackup::RemoveOrder(type, destination, hangar);
}
