/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file order_list.c Handling of orderlist functions. */





#include "../stdafx.h"

#include "orderlist.h"
#include "order_func_local.h"
#include "order_func_global.h"
#include "vehicle_orders.h"

#include "core/pool_func.hpp"

#include "facilities/station_base.h"

#include "vehicle/vehicle_base.h"
#include "vehicle/vehicle_func.h"
#include "vehicle/vehicle_gui.h"





OrderListPool _orderlist_pool("OrderList");
INSTANTIATE_POOL_METHODS(OrderList)





/**
 * Create an order list with the given order chain for the given vehicle.
 *  @param chain pointer to the first order of the order chain
 *  @param v any vehicle using this orderlist
 */

void CreateOrderList (
	OrderList *order_list, Order *chain, Vehicle *vehicle
) {

	InitializeOrderList (order_list, chain, vehicle);
}





/**
 * Recomputes everything.
 * @param chain first order in the chain
 * @param v one of vehicle that is using this orderlist
 */

void InitializeOrderList (
	OrderList *order_list, Order *chain, Vehicle *vehicle
) {

	order_list->first               = chain;
	order_list->first_shared        = vehicle;
	order_list->num_orders          = 0;
	order_list->num_vehicles        = 1;
	order_list->timetable_duration  = 0;

	for (Order *o = order_list->first; o != nullptr; o = o->next) {

		++order_list->num_orders;

		order_list->timetable_duration += o->wait_time + o->travel_time;
	}

	RecalculateTimetableDuration (order_list);

	for (Vehicle *next_vehicle = order_list->first_shared->PreviousShared();
		 next_vehicle != nullptr;
		 next_vehicle = next_vehicle->PreviousShared()
	) {

		++order_list->num_vehicles;

		order_list->first_shared = next_vehicle;
	}

	for (const Vehicle *next_vehicle = vehicle->NextShared();
		 next_vehicle != nullptr;
		 next_vehicle = next_vehicle->NextShared()
	) {

		++order_list->num_vehicles;
	}
}





/**
 * Recomputes Timetable duration.
 * Split out into a separate function so it can be used by afterload.
 */

void RecalculateTimetableDuration (
	OrderList *order_list
) {

	order_list->timetable_duration = 0;

	for (Order *order = order_list->first; order != nullptr; order = order->next) {

		order_list->timetable_duration += order->wait_time + order->travel_time;
	}
}





/**
 * Free a complete order chain.
 * @param keep_orderlist If this is true only delete the orders, otherwise also delete the OrderList.
 * @note do not use on "current_order" vehicle orders!
 */

void FreeChain (
	OrderList *order_list, bool keep_orderlist
) {

	Order *next;

	for (Order *o = order_list->first; o != nullptr; o = next) {

		next = o->next;

		delete o;
	}

	if (keep_orderlist) {

		order_list->first = nullptr;
		order_list->num_orders = 0;
		order_list->timetable_duration = 0;

	} else {

		delete order_list;
	}
}





/**
 * Get a certain order of the order chain.
 * @param index zero-based index of the order within the chain.
 * @return the order at position index.
 */

Order *GetOrderAt(
	const OrderList *order_list, int index
) {

	if (index < 0) {

		return nullptr;
	}

	Order *order = order_list->first;

	while (order != nullptr && index-- > 0) {

		order = order->next;
	}

	return order;
}





/**
 * Move an order to another position within the order list.
 * @param from is the zero-based position of the order to move.
 * @param to is the zero-based position where the order is moved to.
 */

void MoveOrder (
	OrderList *order_list, int from, int to
) {

	if (from >= order_list->num_orders || to >= order_list->num_orders || from == to) return;

	Order *moving_one;

	/* Take the moving order out of the pointer-chain */
	if (from == 0) {
		moving_one = order_list->first;
		order_list->first = moving_one->next;
	} else {
		//check order_list as correct list
		Order *one_before = GetOrderAt (order_list, from - 1);
		moving_one = one_before->next;
		one_before->next = moving_one->next;
	}

	/* Insert the moving_order again in the pointer-chain */
	if (to == 0) {
		moving_one->next = order_list->first;
		order_list->first = moving_one;
	} else {
		// check order_list
		Order *one_before = GetOrderAt (order_list, to - 1);
		moving_one->next = one_before->next;
		one_before->next = moving_one;
	}
}





#ifdef WITH_ASSERT
/**
 * Checks for internal consistency of order list. Triggers assertion if something is wrong.
 */

void OrderListCheckSanity (
	OrderList *order_list
) {

	VehicleOrderID check_num_orders   = 0;
	uint check_num_vehicles           = 0;
	TimeTick check_timetable_duration = 0;

	Debug(misc, 6, "Checking OrderList {} for sanity...", this->index);

	for (const Order *order = order_list->first; order != nullptr; order = order->next) {

		++check_num_orders;

		check_timetable_duration += order->wait_time + order->travel_time;
	}

	assert (order_list->num_orders         == check_num_orders);
	assert (order_list->timetable_duration == check_timetable_duration);

	for (const Vehicle *vehicle = order_list->first_shared; vehicle != nullptr; vehicle = vehicle->NextShared()) {

		++check_num_vehicles;

		assert (vehicle->order_list == order_list);
	}

	assert (order_list->num_vehicles == check_num_vehicles);

	Debug (
			misc, 6,
			"... detected {} orders, {} vehicles, {} timetabled",
			(uint)order_list->num_orders,
			order_list->num_vehicles,
			order_list->timetable_duration
	);
}
#endif





/**
 * Insert a new order but skip the validation.
 * @param v       The vehicle to insert the order to.
 * @param new_o   The new order.
 * @param sel_ord The position the order should be inserted at.
 */

void InsertOrder (
	Vehicle *v, Order *new_o, VehicleOrderID sel_ord
) {

	/* Create new order and link in list */
	if (v->order_list == nullptr) {

		v->order_list = new OrderList;
		CreateOrderList (v->order_list, new_o, v);


	} else {

		InsertOrderAt (v, new_o, sel_ord);
	}

		/* If there is added an order before the current one, we need
		 * to update the selected order. We do not change implicit/real order indices though.
		 * If the new order is between the current implicit order and real order, the implicit order will
		 * later skip the inserted order. */
		if (sel_ord <= v->current_order_index) {
			uint cur = v->current_order_index + 1;
			/* Check if we don't go out of bound */
			if (cur < GetNumOrders (v)) {
				v->current_order_index = cur;
			}
		}

		/* Update any possible open window of the vehicle */
		InvalidateVehicleOrder(v, INVALID_VEH_ORDER_ID | (sel_ord << 8));

	/* As we insert an order, the order to skip to will be 'wrong'. */
	VehicleOrderID cur_order_id = 0;
	for (Order *order : VehicleOrders (v)) {
		if (order->IsType(OT_CONDITIONAL)) {
			VehicleOrderID order_id = order->GetConditionSkipToOrder();
			if (order_id >= sel_ord) {
				order->SetConditionSkipToOrder(order_id + 1);
			}
			if (order_id == cur_order_id) {
				order->SetConditionSkipToOrder((order_id + 1) % GetNumOrders (v));
			}
		}
		cur_order_id++;
	}

	/* Make sure to rebuild the whole list */
	InvalidateWindowClassesData(GetWindowClassForVehicleType(v->type), 0);
}





/**
 * Insert a new order into the order chain.
 * @param new_order is the order to insert into the chain.
 * @param index is the position where the order is supposed to be inserted.
 */

void InsertOrderAt (
	Vehicle *v, Order *new_order, int index
) {

	if (v->order_list->first == nullptr) {

		v->order_list->first = new_order;

	} else {

		if (index == 0) {
			/* Insert as first or only order */
			new_order->next = v->order_list->first;
			v->order_list->first = new_order;

		} else if (index >= v->order_list->num_orders) {

			/* index is after the last order, add it to the end */
			GetOrderAt (v->order_list, v->order_list->num_orders - 1)->next = new_order;

		} else {

			/* Put the new order in between */
			Order *order = GetOrderAt (v->order_list, index - 1);
			new_order->next = order->next;
			order->next = new_order;
		}
	}

	++v->order_list->num_orders;
//	v->order_list->timetable_duration += new_order->wait_time + new_order->travel_time;

	/* We can visit oil rigs and buoys that are not our own. They will be shown in
	 * the list of stations. So, we need to invalidate that window if needed. */
	if (new_order->IsType(OT_GOTO_STATION) || new_order->IsType(OT_GOTO_WAYPOINT)) {

		BaseStation *bs = BaseStation::Get(new_order->dest);

		if (bs->owner == OWNER_NONE) InvalidateWindowClassesData(WC_STATION_LIST, 0);
	}

}





/**
 * Delete an order but skip the parameter validation.
 * @param v       The vehicle to delete the order from.
 * @param sel_ord The id of the order to be deleted.
 */

void DeleteOrder (
	Vehicle *v, VehicleOrderID sel_ord
) {

	DeleteOrderAt (v->order_list, v, sel_ord);

	DeleteOrderWarnings(v);

		if (sel_ord == v->current_order_index && v->current_order.IsType(OT_LOADING)) {
			CancelLoadingDueToDeletedOrder(v);
		}

		if (sel_ord < v->current_order_index) {
			--v->current_order_index;
		} else if (sel_ord == v->current_order_index) {
			ToNextOrderIndex(v);
		}

		/* Update any possible open window of the vehicle */
		InvalidateVehicleOrder(v, sel_ord | (INVALID_VEH_ORDER_ID << 8));

	/* As we delete an order, the order to skip to will be 'wrong'. */
	VehicleOrderID cur_order_id = 0;
	for (Order *order : VehicleOrders (v)) {
		if (order->IsType(OT_CONDITIONAL)) {
			VehicleOrderID order_id = order->GetConditionSkipToOrder();
			if (order_id >= sel_ord) {
				order_id = std::max(order_id - 1, 0);
			}
			if (order_id == cur_order_id) {
				order_id = (order_id + 1) % GetNumOrders (v);
			}
			order->SetConditionSkipToOrder(order_id);
		}
		++cur_order_id;
	}

	InvalidateWindowClassesData(GetWindowClassForVehicleType(v->type), 0);
}





/**
 * Remove an order from the order list and delete it.
 * @param index is the position of the order which is to be deleted.
 */

void DeleteOrderAt (
	OrderList *order_list, Vehicle *v, int index
) {

	if (index >= v->order_list->num_orders) return;

	Order *to_remove;

	if (index == 0) {

		to_remove = v->order_list->first;
		v->order_list->first = to_remove->next;

	} else {

		//check this as correct list
		Order *prev = GetOrderAt (order_list, index - 1);
		to_remove = prev->next;
		prev->next = to_remove->next;
	}

	--order_list->num_orders;

	v->order_list->timetable_duration -= (to_remove->wait_time + to_remove->travel_time);

	delete to_remove;
}





/**
 * Get the order after the given one or the first one, if the given one is the
 * last one.
 * @param curr Order to find the next one for.
 * @return Next order.
 */

const Order *GetNextOrder (
	const OrderList *order_list, const Order *curr
) {

	return (curr->next == nullptr) ? order_list->first : curr->next;
}
