/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file vehicle_orders.h Functions relying both on vehicle and order objects. */





#include "vehicle_orders.h"

#include "orderlist.h"
#include "order_func_local.h"
#include "order_func_global.h"

#include "facilities/depot_base.h"
#include "facilities/station_base.h"

#include "game/cheat_func.h"

#include "vehicle/aircraft.h"
#include "vehicle/vehicle_func.h"
#include "vehicle/vehicle_gui.h"
#include "vehicle/vehicle_logic.h"

#include "news_func.h"
#include "strings_func.h"

#include "debug.h"






/**
 * Handle the orders of a vehicle and determine the next place
 * to go to if needed.
 * @param v the vehicle to do this for.
 * @return true *if* the vehicle is eligible for reversing
 *              (basically only when leaving a station).
 */

bool ProcessOrders (
	Vehicle *vehicle
) {

	/* Get the current order */
	Order *order = GetVehicleOrder (vehicle, vehicle->current_order_index);
	if (order == nullptr) {
		return false;
	}

	switch (vehicle->current_order.GetType()) {
		case OT_GOTO_DEPOT: {
			/* Let a depot order in the orderlist interrupt. */
			if (!(vehicle->current_order.GetDepotOrderType() & ODTFB_PART_OF_ORDERS)) return false;
			break;
		}

		case OT_LOADING:
		case OT_LEAVESTATION: {
			return false;
		}

		case OT_GOTO_STATION: {
			break;
		}

		case OT_GOTO_WAYPOINT: {
			if (vehicle->vehicle_location & VL_AT_WAYPOINT) return false;
			break;
		}

		default: break;
	}

	/* If no order, do nothing. */
	if ((
			vehicle->type == VEH_AIRCRAFT
			&& !CheckForValidOrders(vehicle)
		)
	) {
		if (vehicle->type == VEH_AIRCRAFT) {
			/* Aircraft do something vastly different here, so handle separately */
			extern void HandleMissingAircraftOrders(Aircraft *vehicle);
			HandleMissingAircraftOrders(Aircraft::From(vehicle));

			return false;
		}

		vehicle->current_order.Free();
		vehicle->SetDestTile(0);
		return false;
	}

	/* If it is unchanged, keep it. */
	if (
		order->Equals(vehicle->current_order)
		&& (
			vehicle->type == VEH_AIRCRAFT
			|| vehicle->dest_tile != 0
		)
		&& (
			vehicle->type != VEH_SHIP
			|| !order->IsType(OT_GOTO_STATION)
			|| !order->IsType(OT_GOTO_WAYPOINT)
			|| Station::Get(order->dest)->ship_station.tile != INVALID_TILE
		)
	) {
		return false;
	}

	/* Otherwise set it, and determine the destination tile. */
	vehicle->current_order = *order;

	InvalidateVehicleOrder(vehicle, VIWD_MODIFY_ORDERS);

	switch (vehicle->type) {

		case VEH_ROAD:
		case VEH_TRAIN:
			break;

		case VEH_AIRCRAFT:
		case VEH_SHIP:
			SetWindowClassesDirty(GetWindowClassForVehicleType(vehicle->type));
			break;

		default:
			NOT_REACHED();
	}

	/* @todo may_reverse obsolete with removal of OT_NOTHING orders,
	 * replace with other/new flag trigger.
	*/
//	return UpdateVehicleDestination (vehicle, order);
}





/**
 * Check whether the given vehicle should stop at the given station
 * based on this order and the non-stop settings.
 * @param v       the vehicle that might be stopping.
 * @param station the station to stop at.
 * @return true if the vehicle should stop.
 */

bool ShouldStopAtStation (
	Vehicle *vehicle
) {

	if ( vehicle->current_order.dest != GetStationIndex (vehicle->tile) ) {

		return false;
	}

	bool stop            = false;
	bool waypoint        = vehicle->tile == INVALID_TILE
						   ? false
						   : GetStationType (vehicle->tile) == STATION_WAYPOINT
	;
	Order order          = vehicle->current_order;
	StationID station_id = GetStationIndex (vehicle->tile);

	if (waypoint) {

		stop = order.IsType (OT_GOTO_WAYPOINT)
				&& order.dest == station_id
				&& (
					   order.wait_time           > 0
					|| order.scheduled_departure > 0
					|| (order.GetWaypointFlags() & OWF_REVERSE)
				)
		;

		if (stop)  vehicle->vehicle_location |= VL_ENTER_WAYPOINT;
		if (!stop) vehicle->vehicle_location |= VL_SKIP_LOCATION;
	}

	if (order.IsType (OT_GOTO_STATION)) {

Debug (runtime, 3, "stop at station?");
		stop = order.dest == station_id
				&& (
					   (!(order.GetLoadType()    & OLFB_NO_LOAD))
					|| (!(order.GetUnloadType()  & OUFB_NO_UNLOAD))
					|| order.wait_time           > 0
					|| order.scheduled_departure > 0
				)
		;

		if (stop)  vehicle->vehicle_location |= VL_AT_STATION;
		if (!stop) vehicle->vehicle_location |= VL_SKIP_LOCATION;
	}

	if (order.IsType (OT_GOTO_DEPOT) ) {

		stop = (
				   order.GetDepotOrderType() & ODTFB_PART_OF_ORDERS
				&& IsDepotTile (vehicle->tile)
			   )
	;

		if (stop)  vehicle->vehicle_location |= VL_ENTER_DEPOT;
		if (!stop) vehicle->vehicle_location |= VL_SKIP_LOCATION;
	}

Debug (runtime, 3, "should stop {}", stop);
	return stop;
}





/**
 * Removes the vehicle from the shared order list.
 * @note This is supposed to be called when the vehicle is still in the chain
 * @param v vehicle to remove from the list
 */

void RemoveVehicle (
	Vehicle *vehicle
) {

	--vehicle->order_list->num_vehicles;

	if (vehicle == vehicle->order_list->first_shared) {

		vehicle->order_list->first_shared = vehicle->NextShared();
	}
}





/**
 * Gets the position of the given vehicle within the shared order vehicle list.
 * @param v is the vehicle of which to get the position
 * @return position of v within the shared vehicle chain.
 */

int GetPositionInSharedOrderList (
	const Vehicle *vehicle
) {

	int count = 0;

	for (
			const Vehicle *vehicle_shared  = vehicle->PreviousShared();
			vehicle_shared                != nullptr;
			vehicle_shared                 = vehicle_shared->PreviousShared()
	) {

		count++;
	}

	return count;
}





/**
 * Returns a tile somewhat representing the order destination (not suitable for pathfinding).
 * @param v The vehicle to get the location for.
 * @param airport Get the airport tile and not the station location for aircraft.
 * @return destination of order, or INVALID_TILE if none.
 */

TileIndex GetLocation (
	const Vehicle *vehicle, const Order *order, bool airport
) {

	switch (order->GetType()) {
		case OT_GOTO_WAYPOINT:
		case OT_GOTO_STATION:
			if (airport && vehicle->type == VEH_AIRCRAFT) return Station::Get(order->dest)->airport.tile;
			return BaseStation::Get(order->dest)->xy;

		case OT_GOTO_DEPOT:
			if ((order->GetDepotActionType() & ODATFB_NEAREST_DEPOT) != 0) return INVALID_TILE;
			return (vehicle->type == VEH_AIRCRAFT) ? Station::Get(order->dest)->xy : Depot::Get(order->dest)->xy;

		default:
			return INVALID_TILE;
	}
}





/**
 * Get the next order which will make the given vehicle stop at a station
 * or refit at a depot or evaluate a non-trivial condition.
 * @param next The order to start looking at.
 * @param hops The number of orders we have already looked at.
 * @return Either of
 *         \li a station order
 *         \li a refitting depot order
 *         \li a non-trivial conditional order
 *         \li nullptr  if the vehicle won't stop anymore.
 */

const Order *GetNextDecisionNode (
	const OrderList *order_list, const Order *next, uint hops
) {

	if (hops > order_list->num_orders || next == nullptr) {

		return nullptr;
	}

	if (next->IsType(OT_CONDITIONAL)) {

		if (next->GetConditionVariable() != OCV_UNCONDITIONALLY) {

			return next;
		}

		/* We can evaluate trivial conditions right away. They're conceptually
		 * the same as regular order progression. */
		return GetNextDecisionNode (order_list,
									GetOrderAt (order_list,
												next->GetConditionSkipToOrder()
											   ),
									hops + 1
								   )
		;
	}

	if (next->IsType(OT_GOTO_DEPOT)) {

		if (next->GetDepotActionType() == ODATFB_HALT) return nullptr;
		if (next->GetDepotActionType() == ODATFB_PARK) return nullptr;
		if (next->IsRefit())                           return next;
	}

	if (!CanLoadOrUnload (next)) {

		return GetNextDecisionNode (order_list, GetNextOrder (order_list, next), hops + 1);
	}

	return next;
}





/**
 * Recursively determine the next deterministic station to stop at.
 * @param v The vehicle we're looking at.
 * @param first Order to start searching at or nullptr to start at cur_implicit_order_index + 1.
 * @param hops Number of orders we have already looked at.
 * @return Next stopping station or INVALID_STATION.
 * @pre The vehicle is currently loading and v->last_station_visited is meaningful.
 * @note This function may draw a random number. Don't use it from the GUI.
 */

StationIDStack GetNextStoppingStationOrder (
	OrderList *order_list, const Vehicle *v, const Order *first, uint hops
) {

	const Order *next = first;

	if (first == nullptr) {

		next = GetOrderAt (order_list, v->current_order_index);

		if (next == nullptr) {

			next = order_list->first;

			if (next == nullptr) return INVALID_STATION;

		} else {

			/* GetNext never returns nullptr if there is a valid station in the list.
			 * As the given "next" is already valid and a station in the list, we
			 * don't have to check for nullptr here. */
			next = GetNextOrder (order_list, next);
			assert(next != nullptr);
		}
	}

	do {
		next = GetNextDecisionNode (order_list, next, ++hops);

		/* Resolve possibly nested conditionals by estimation. */

		while (next != nullptr && next->IsType(OT_CONDITIONAL)) {

			/* We return both options of conditional orders. */
			const Order *skip_to = GetNextDecisionNode (order_list,
														GetOrderAt (order_list,
																	next->GetConditionSkipToOrder()
																   ),
													   hops
													   )
			;

			const Order *advance = GetNextDecisionNode (order_list,
														GetNextOrder (order_list, next),
														hops
													   )
			;

			if (advance == nullptr || advance == first || skip_to == advance) {

				next = (skip_to == first) ? nullptr : skip_to;

			} else if (skip_to == nullptr || skip_to == first) {

				next = (advance == first) ? nullptr : advance;

			} else {

				StationIDStack st1 = GetNextStoppingStationOrder (order_list, v, skip_to, hops);
				StationIDStack st2 = GetNextStoppingStationOrder (order_list, v, advance, hops);

				while (!st2.IsEmpty()) st1.Push(st2.Pop());

				return st1;
			}

			++hops;
		}

		/* Don't return a next stop if the vehicle has to unload everything. */
		if (next == nullptr || (next->IsType(OT_GOTO_STATION) &&
				next->dest == v->last_station_visited &&
				(next->GetUnloadType() & (OUFB_TRANSFER | OUFB_UNLOAD)) != 0)) {

			return INVALID_STATION;
		}

	} while (next->IsType(OT_GOTO_DEPOT) || next->dest == v->last_station_visited);

	return next->dest;
}





/**
 * Delete all news items regarding defective orders about a vehicle
 * This could kill still valid warnings (for example about void order when just
 * another order gets added), but assume the company will notice the problems,
 * when they're changing the orders.
 */

void DeleteOrderWarnings (
	const Vehicle *v
) {

	DeleteVehicleNews(v->index, STR_NEWS_VEHICLE_HAS_TOO_FEW_ORDERS);
	DeleteVehicleNews(v->index, STR_NEWS_VEHICLE_HAS_VOID_ORDER);
	DeleteVehicleNews(v->index, STR_NEWS_VEHICLE_HAS_DUPLICATE_ENTRY);
	DeleteVehicleNews(v->index, STR_NEWS_VEHICLE_HAS_INVALID_ENTRY);
	DeleteVehicleNews(v->index, STR_NEWS_PLANE_USES_TOO_SHORT_RUNWAY);
}





/**
 * Get the distance between two orders of a vehicle. Conditional orders are resolved
 * and the bigger distance of the two order branches is returned.
 * @param prev Origin order.
 * @param cur Destination order.
 * @param v The vehicle to get the distance for.
 * @param conditional_depth Internal param for resolving conditional orders.
 * @return Maximum distance between the two orders.
 */

uint GetOrderDistance (
	const Order *prev, const Order *cur, const Vehicle *v, int conditional_depth
) {

	if (cur->IsType(OT_CONDITIONAL)) {

		if (conditional_depth > GetNumOrders (v)) return 0;

		conditional_depth++;

		int dist1 = GetOrderDistance(prev, GetVehicleOrder (v, cur->GetConditionSkipToOrder()), v, conditional_depth);
		int dist2 = GetOrderDistance(prev, cur->next == nullptr ? v->order_list->first : cur->next, v, conditional_depth);

		return std::max(dist1, dist2);
	}

	TileIndex prev_tile = GetLocation(v, prev, true);
	TileIndex cur_tile = GetLocation(v, cur, true);

	if (prev_tile == INVALID_TILE || cur_tile == INVALID_TILE) return 0;

	return v->type == VEH_AIRCRAFT ? DistanceSquare(prev_tile, cur_tile) : DistanceManhattan(prev_tile, cur_tile);
}





/**
 * Cancel the current loading order of the vehicle as the order was deleted.
 * @param v the vehicle
 */
void CancelLoadingDueToDeletedOrder(Vehicle *v)
{
	assert(v->current_order.IsType(OT_LOADING));
	/* NON-stop flag is misused to see if a train is in a station that is
	 * on its order list or not */
	v->current_order.SetNonStopType(ONSF_STOP_EVERYWHERE);
	/* When full loading, "cancel" that order so the vehicle doesn't
	 * stay indefinitely at this station anymore. */
	if (v->current_order.GetLoadType() & OLFB_FULL_LOAD) v->current_order.SetLoadType(OLF_LOAD_IF_POSSIBLE);
}





/**
 * Check if a vehicle has any valid orders
 *
 * @return false if there are no valid orders
 * @note Conditional orders are not considered valid destination orders
 */

bool CheckForValidOrders (
	const Vehicle *v
) {

	for (const Order *order : VehicleOrders (v)) {
		switch (order->GetType()) {
			case OT_GOTO_STATION:
			case OT_GOTO_DEPOT:
			case OT_GOTO_WAYPOINT:
				return true;

			default:
				break;
		}
	}

	return false;
}





/**
 * Check the orders of a vehicle, to see if there are invalid orders and stuff
 */

void CheckVehicleOrders(const Vehicle *v)
{
	/* Does the user wants us to check things? */
	if (_settings_client.gui.order_review_system == 0) return;

	/* Do nothing for crashed vehicles */
	if (v->vehstatus & VS_CRASHED) return;

	/* Do nothing for stopped vehicles if setting is '1' */
	if (_settings_client.gui.order_review_system == 1 && (v->vehstatus & VS_STOPPED)) return;

	/* do nothing we we're not the first vehicle in a share-chain */
	if (v->FirstShared() != v) return;

	/* Only check every 20 days, so that we don't flood the message log */
	if (v->owner == _local_company && v->day_counter % 20 == 0) {
		StringID message = INVALID_STRING_ID;

		/* Check the order list */
		int n_st = 0;

		for (const Order *order : VehicleOrders (v)) {
			/* Dummy order? */
			if (order->IsType(OT_DUMMY)) {
				message = STR_NEWS_VEHICLE_HAS_VOID_ORDER;
				break;
			}
			/* Does station have a load-bay for this vehicle? */
			if (order->IsType(OT_GOTO_STATION)) {
				const Station *st = Station::Get(order->dest);

				n_st++;
				if (!CanVehicleUseStation(v, st)) {
					message = STR_NEWS_VEHICLE_HAS_INVALID_ENTRY;
				} else if (v->type == VEH_AIRCRAFT &&
							(AircraftVehInfo(v->engine_type)->subtype & AIR_FAST) &&
							(st->airport.GetFTA()->flags & AirportFTAClass::SHORT_STRIP) &&
							!_cheats.no_jetcrash.value &&
							message == INVALID_STRING_ID) {
					message = STR_NEWS_PLANE_USES_TOO_SHORT_RUNWAY;
				}
			}
		}

		/* Check if the last and the first order are the same */
		if (GetNumOrders (v) > 1) {
			const Order *last = GetOrderAt (v->order_list, v->order_list->num_orders - 1);

			if (v->order_list->first->Equals(*last)) {
				message = STR_NEWS_VEHICLE_HAS_DUPLICATE_ENTRY;
			}
		}

		/* Do we only have 1 station in our order list? */
		if (n_st < 2 && message == INVALID_STRING_ID) message = STR_NEWS_VEHICLE_HAS_TOO_FEW_ORDERS;

#ifdef WITH_ASSERT
		if (v->order_list != nullptr) OrderListCheckSanity (v_>order_list);
#endif

		/* We don't have a problem */
		if (message == INVALID_STRING_ID) return;

		SetDParam(0, v->index);
		AddVehicleAdviceNewsItem(message, v->index);
	}
}





/**
 * Delete all orders from a vehicle
 * @param v                   Vehicle whose orders to reset
 * @param keep_orderlist      If true, do not free the order list, only empty it.
 * @param reset_order_indices If true, reset cur_implicit_order_index and current_order_index
 *                            and cancel the current full load order (if the vehicle is loading).
 *                            If false, _you_ have to make sure the order indices are valid after
 *                            your messing with them!
 */

void DeleteVehicleOrders (
	Vehicle *v, bool keep_orderlist, bool reset_order_indices
) {

	DeleteOrderWarnings(v);

	if (v->vehicle_flags & VF_SHARED_ORDERS) {

		/* Remove ourself from the shared order list. */
		RemoveFromShared(v);
		v->order_list = nullptr;

	} else if (v->order_list != nullptr) {

		/* Remove the orders */
		FreeChain (v->order_list, keep_orderlist);
		if (!keep_orderlist) v->order_list = nullptr;
	}

	if (reset_order_indices) {

		v->current_order_index = 0;

		if (v->current_order.IsType(OT_LOADING)) {

			CancelLoadingDueToDeletedOrder(v);
		}
	}
}
