/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file  vehicle_base.h Base class for all vehicles. */



#ifndef VEHICLE_BASE_H
#define VEHICLE_BASE_H



#include <list>
#include <map>

#include "base_consist.h"
#include "vehicle_type.h"

#include "engine_type.h"
#include "order_iterator.h"

#include "../cargo/cargopacket.h"

#include "../core/enum_type.hpp"
#include "../core/pool_type.hpp"
#include "../core/smallmap_type.hpp"

#include "../infrastructure/track_type.h"
#include "../logic/group_type.h"

#include "../network/network.h"

#include "../orders/order.h"
#include "../orders/orderlist.h"
#include "../orders/timetable.h"

#include "../saveload/saveload.h"

#include "../command_type.h"
#include "../texteff.hpp"
#include "../transport_type.h"





/** Cached often queried values common to all vehicles. */
struct VehicleCache {
	uint16 cached_max_speed;          ///< Maximum speed of the consist
									  ///< (minimum of the max speed of all vehicles in the consist).
	TimeTick cached_cargo_age_period; ///< Number of ticks before carried cargo is aged.
	byte cached_vis_effect;  		  ///< Visual effect to show (see #VisualEffect)
};





/** Sprite sequence for a vehicle part. */
struct VehicleSpriteSeq {
	PalSpriteID seq[4];
	uint count;

	bool operator==(const VehicleSpriteSeq &other) const
	{
		return this->count == other.count && MemCmpT<PalSpriteID>(this->seq, other.seq, this->count) == 0;
	}

	bool operator!=(const VehicleSpriteSeq &other) const
	{
		return !this->operator==(other);
	}

	/**
	 * Check whether the sequence contains any sprites.
	 */
	bool IsValid() const
	{
		return this->count != 0;
	}

	/**
	 * Clear all information.
	 */
	void Clear()
	{
		this->count = 0;
	}

	/**
	 * Assign a single sprite to the sequence.
	 */
	void Set(SpriteID sprite)
	{
		this->count = 1;
		this->seq[0].sprite = sprite;
		this->seq[0].pal = 0;
	}

	/**
	 * Copy data from another sprite sequence, while dropping all recolouring information.
	 */
	void CopyWithoutPalette(const VehicleSpriteSeq &src)
	{
		this->count = src.count;
		for (uint i = 0; i < src.count; ++i) {
			this->seq[i].sprite = src.seq[i].sprite;
			this->seq[i].pal = 0;
		}
	}

	void GetBounds(Rect *bounds) const;
	void Draw(int x, int y, PaletteID default_pal, bool force_pal) const;
};





/**
 * Cache for vehicle sprites and values relating to whether they should be updated before drawing,
 * or calculating the viewport.
 */
struct MutableSpriteCache {
	Direction last_direction;     ///< Last direction we obtained sprites for
	bool revalidate_before_draw;  ///< We need to do a GetImage() and check bounds before drawing this sprite
	Rect old_coord;               ///< Co-ordinates from the last valid bounding box
	bool is_viewport_candidate;   ///< This vehicle can potentially be drawn on a viewport
	VehicleSpriteSeq sprite_seq;  ///< Vehicle appearance.
};





/**
 * Simulated cargo type and capacity for prediction of future links.
 */
struct RefitDesc {
	CargoID cargo;    ///< Cargo type the vehicle will be carrying.
	uint16 capacity;  ///< Capacity the vehicle will have.
	uint16 remaining; ///< Capacity remaining from before the previous refit.
	RefitDesc(CargoID cargo, uint16 capacity, uint16 remaining) :
			cargo(cargo), capacity(capacity), remaining(remaining) {}
};





/** %Vehicle data structure. */
struct Vehicle : VehiclePool::PoolItem<&_vehicle_pool>, BaseVehicle, BaseConsist {
//private:
//public:

	friend void   FixOldVehicles     ();
	friend void   AfterLoadVehicles  (bool part_of_load);           ///< So we can set the #previous and #first pointers while loading
	friend bool   LoadOldVehicle     (LoadgameState *ls, int num);  ///< So we can set the proper next pointer while loading

	/* So we can use private/protected variables in the saveload code */
	friend class  SlVehicleCommon;
	friend class  SlVehicleDisaster;
	friend void   Ptrs_VEHS          ();



	typedef std::list<RefitDesc> RefitList;
	typedef std::map<CargoID, uint> CapacitiesMap;



	Vehicle   *next;                      ///< pointer to the next vehicle in the chain
	Vehicle   *previous;                  ///< NOSAVE: pointer to the previous vehicle in the chain
	Vehicle   *first;                     ///< NOSAVE: pointer to the first vehicle in the chain

	Vehicle   *next_shared;               ///< pointer to the next vehicle that shares the order
	Vehicle   *previous_shared;           ///< NOSAVE: pointer to the previous vehicle in the shared order chain

	TileIndex tile;                       ///< Current tile index
	TileIndex previous_tile;              ///< Previous tile index

	/**
	 * Heading for this tile.
	 * For airports and train stations this tile does not necessarily belong to the destination station,
	 * but it can be used for heuristic purposes to estimate the distance.
	 */
	TileIndex dest_tile;

	Money profit_this_year;             ///< Profit this year << 8, low 8 bits are fract
	Money profit_last_year;             ///< Profit last year << 8, low 8 bits are fract
	Money value;                        ///< Value of the vehicle

	CargoPayment *cargo_payment;        ///< The cargo payment we're currently in

	mutable   Rect coord;                 ///< NOSAVE: Graphical bounding box of the vehicle, i.e. what to redraw on moves.

	Vehicle  *hash_viewport_next;        ///< NOSAVE: Next vehicle in the visual location hash.
	Vehicle **hash_viewport_prev;       ///< NOSAVE: Previous vehicle in the visual location hash.

	Vehicle  *hash_tile_next;            ///< NOSAVE: Next vehicle in the tile location hash.
	Vehicle **hash_tile_prev;           ///< NOSAVE: Previous vehicle in the tile location hash.
	Vehicle **hash_tile_current;        ///< NOSAVE: Cache of the current hash chain.

	SpriteID colourmap;                 ///< NOSAVE: cached colour mapping

	/* Related to age and service time */
	Year      build_year;                     ///< Year the vehicle has been built.
	Date      age;                            ///< Age in days
	Date      max_age;                        ///< Maximum age
	Date      date_of_last_service;           ///< Last date the vehicle had a service at a depot.
	uint16    reliability;                    ///< Reliability.
	uint16    reliability_spd_dec;            ///< Reliability decrease speed.
	TimeTick  service_time;                   ///< Time needed for servicing the vehicle.
	byte      breakdown_ctr;                  ///< Counter for managing breakdown events. @see HandleBreakdown
	byte      breakdown_delay;                ///< Counter for managing breakdown length.
	byte      breakdowns_since_last_service;  ///< Counter for the amount of breakdowns.
	byte      breakdown_chance;               ///< Current chance of breakdowns.

	int32      x_pos;                         ///< x coordinate.
	int32      y_pos;                         ///< y coordinate.
	int32      z_pos;                         ///< z coordinate.
	Direction  direction;                     ///< facing

	Owner owner;                              ///< Which company owns the vehicle?

	/**
	 * currently displayed sprite index
	 * 0xfd == custom sprite, 0xfe == custom second head sprite
	 * 0xff == reserved for another custom sprite
	 */
	byte spritenum;
	byte x_extent;                      ///< x-extent of vehicle bounding box
	byte y_extent;                      ///< y-extent of vehicle bounding box
	byte z_extent;                      ///< z-extent of vehicle bounding box
	int8 x_bb_offs;                     ///< x offset of vehicle bounding box
	int8 y_bb_offs;                     ///< y offset of vehicle bounding box
	int8 x_offs;                        ///< x offset for vehicle sprite
	int8 y_offs;                        ///< y offset for vehicle sprite

	EngineID engine_type;               ///< The type of engine used for this vehicle.

	TextEffectID fill_percent_te_id;    ///< a text-effect id to a loading indicator object
	UnitID unitnumber;                  ///< unit number, for display purposes only

	uint16 cur_speed;                   ///< current speed
	byte subspeed;                      ///< fractional speed
	byte acceleration;                  ///< used by train & aircraft
	uint32 motion_counter;              ///< counter to occasionally play a vehicle sound.
	byte progress;                      ///< The percentage (if divided by 256) this vehicle already crossed the tile unit.

	byte random_bits;                   ///< Bits used for determining which randomized variational spritegroups to use when drawing.
	byte waiting_triggers;              ///< Triggers to be yet matched before rerandomizing the random bits.

	StationID last_station_visited;     ///< The last station we stopped at.
	StationID last_loading_station;     ///< Last station the vehicle has stopped at and could possibly leave from with any cargo loaded.

	CargoID cargo_type;                 ///< type of cargo this vehicle is carrying
	byte cargo_subtype;                 ///< Used for livery refits (NewGRF variations)
	uint16 cargo_cap;                   ///< total capacity
	uint16 refit_cap;                   ///< Capacity left over from before last refit.
	VehicleCargoList cargo;             ///< The cargo this vehicle is carrying
	TimeTick cargo_age_counter;         ///< Ticks till cargo is aged next.
	int8 trip_occupancy;                ///< NOSAVE: Occupancy of vehicle of the current trip (updated after leaving a station).

	byte day_counter;                   ///< Increased by one for each day
	TimeTick tick_counter;              ///< Increased by one for each tick
	TimeTick running_ticks;             ///< Number of ticks this vehicle was not stopped this day

	int64 vehstatus;                    ///< Vehicle status
	int64 vehicle_location;             ///< Vehicle location
	int64 vehicle_action;               ///< Vehicle action

	Order current_order;                ///< The current order (+ status, like: loading)

	OrderList *order_list = nullptr;

	TimeTick load_unload_ticks;         ///< Ticks to wait before starting next cycle.

	GroupID group_id;                   ///< Index of group Pool array
	byte subtype;                       ///< subtype (Filled with values from #AircraftSubType/#DisasterSubType/#EffectVehicleType
										///<									/#GroundVehicleSubtypeFlags)

	NewGRFCache grf_cache;              ///< Cache of often used calculated NewGRF values
	VehicleCache vcache;                ///< Cache of often used vehicle values.

	mutable MutableSpriteCache sprite_cache; ///< Cache of sprites and values related to recalculating them, see #MutableSpriteCache



	Vehicle (VehicleType type = VEH_INVALID);



	void PreDestructor();
	/** We want to 'destruct' the right class. */
	virtual ~Vehicle();



	GroundVehicleCache *GetGroundVehicleCache();
	const GroundVehicleCache *GetGroundVehicleCache() const;

	virtual void MarkDirty();
	virtual void UpdateDeltaXY();
	uint GetOldAdvanceSpeed(uint speed);
	static uint GetAdvanceSpeed(uint speed);
	uint GetAdvanceDistance();
	virtual ExpensesType GetExpenseType(bool income) const;
	virtual void PlayLeaveStationSound() const;
	virtual bool IsPrimaryVehicle() const;

	virtual void GetImage(Direction direction, EngineImageType image_type, VehicleSpriteSeq *result) const;

	void InvalidateNewGRFCache();
	void InvalidateNewGRFCacheOfChain();
	bool IsGroundVehicle() const;
	virtual int GetDisplaySpeed() const;
	virtual int GetDisplayMaxSpeed() const;
	virtual int GetCurrentMaxSpeed();
	virtual Money GetRunningCost() const;
	virtual bool IsInDepot() const;
	virtual bool IsChainInDepot() const;
	bool IsStoppedInDepot() const;
	virtual bool Tick();
	virtual void OnNewDay() {};

	virtual uint Crash(bool flooded);
	virtual Trackdir GetVehicleTrackdir() const;
	Money GetDisplayRunningCost() const;
	Money GetDisplayProfitThisYear() const;
	Money GetDisplayProfitLastYear() const;

	void SetNext(Vehicle *next);

	Vehicle *Next() const;
	Vehicle *Previous() const;
	Vehicle *First() const;
	Vehicle *Last();
	const Vehicle *Last() const;
	Vehicle *Move(int n);
	const Vehicle *Move(int n) const;
	Vehicle *NextShared() const;
	Vehicle *PreviousShared() const;
	Vehicle *FirstShared() const;
	bool IsOrderListShared() const;
	void ResetRefitCaps();
	void CopyVehicleConfigAndStatistics(const Vehicle *src);

	bool NeedsAutorenewing(const Company *c, bool use_renew_setting = true) const;

	bool NeedsServicing() const;
	bool NeedsAutomaticServicing() const;

	virtual TileIndex GetOrderStationLocation(StationID station);
	virtual bool FindClosestDepot(TileIndex *location, DestinationID *destination, bool *reverse);

	virtual void SetDestTile(TileIndex tile);

	CommandCost SendToDepot(ProcessCmdFlag flags, DepotCommand command);

	void UpdateVisualEffect(bool allow_power_change);
	void ShowVisualEffect();

	void UpdatePosition();
	void UpdateViewport(bool dirty);
	void UpdateBoundingBoxCoordinates(bool update_cache) const;
	void UpdatePositionAndViewport();
	bool MarkAllViewportsDirty() const;

	uint16 GetServiceInterval() const;

	void SetServiceInterval(uint16 interval);

	bool ServiceIntervalIsCustom() const;

	bool ServiceIntervalIsPercent() const;

	void SetServiceIntervalIsCustom(bool on);

	void SetServiceIntervalIsPercent(bool on);

	bool IsFrontEngine() const;
	bool IsArticulatedPart() const;
	bool HasArticulatedPart() const;
	Vehicle *GetNextArticulatedPart() const;
	Vehicle *GetFirstEnginePart();
	const Vehicle *GetFirstEnginePart() const;
	Vehicle *GetLastEnginePart();
	Vehicle *GetNextVehicle() const;
	Vehicle *GetPrevVehicle() const;
};





/**
 * Class defining several overloaded accessors so we don't
 * have to cast vehicle types that often
 */
template <class T, VehicleType Type>
struct SpecializedVehicle : public Vehicle {
	static const VehicleType EXPECTED_TYPE = Type; ///< Specialized type

	typedef SpecializedVehicle<T, Type> SpecializedVehicleBase; ///< Our type

	/**
	 * Set vehicle type correctly
	 */
	SpecializedVehicle<T, Type>() : Vehicle(Type)
	{
		this->sprite_cache.sprite_seq.count = 1;
	}

	/**
	 * Get the first vehicle in the chain
	 * @return first vehicle in the chain
	 */
	T *First() const { return (T *)this->Vehicle::First(); }

	/**
	 * Get the last vehicle in the chain
	 * @return last vehicle in the chain
	 */
	T *Last() { return (T *)this->Vehicle::Last(); }

	/**
	 * Get the last vehicle in the chain
	 * @return last vehicle in the chain
	 */
	const T *Last() const { return (const T *)this->Vehicle::Last(); }

	/**
	 * Get next vehicle in the chain
	 * @return next vehicle in the chain
	 */
	T *Next() const { return (T *)this->Vehicle::Next(); }

	/**
	 * Get previous vehicle in the chain
	 * @return previous vehicle in the chain
	 */
	T *Previous() const { return (T *)this->Vehicle::Previous(); }

	/**
	 * Get the next part of an articulated engine.
	 * @return Next part of the articulated engine.
	 * @pre The vehicle is an articulated engine.
	 */
	T *GetNextArticulatedPart() { return (T *)this->Vehicle::GetNextArticulatedPart(); }

	/**
	 * Get the next part of an articulated engine.
	 * @return Next part of the articulated engine.
	 * @pre The vehicle is an articulated engine.
	 */
	T *GetNextArticulatedPart() const { return (T *)this->Vehicle::GetNextArticulatedPart(); }

	/**
	 * Get the first part of an articulated engine.
	 * @return First part of the engine.
	 */
	T *GetFirstEnginePart() { return (T *)this->Vehicle::GetFirstEnginePart(); }

	/**
	 * Get the first part of an articulated engine.
	 * @return First part of the engine.
	 */
	const T *GetFirstEnginePart() const { return (const T *)this->Vehicle::GetFirstEnginePart(); }

	/**
	 * Get the last part of an articulated engine.
	 * @return Last part of the engine.
	 */
	T *GetLastEnginePart() { return (T *)this->Vehicle::GetLastEnginePart(); }

	/**
	 * Get the next real (non-articulated part) vehicle in the consist.
	 * @return Next vehicle in the consist.
	 */
	T *GetNextVehicle() const { return (T *)this->Vehicle::GetNextVehicle(); }

	/**
	 * Get the previous real (non-articulated part) vehicle in the consist.
	 * @return Previous vehicle in the consist.
	 */
	T *GetPrevVehicle() const { return (T *)this->Vehicle::GetPrevVehicle(); }

	/**
	 * Tests whether given index is a valid index for vehicle of this type
	 * @param index tested index
	 * @return is this index valid index of T?
	 */
	static bool IsValidID(size_t index)
	{
		return Vehicle::IsValidID(index) && Vehicle::Get(index)->type == Type;
	}

	/**
	 * Gets vehicle with given index
	 * @return pointer to vehicle with given index casted to T *
	 */
	static T *Get(size_t index)
	{
		return (T *)Vehicle::Get(index);
	}

	/**
	 * Returns vehicle if the index is a valid index for this vehicle type
	 * @return pointer to vehicle with given index if it's a vehicle of this type
	 */
	static T *GetIfValid(size_t index)
	{
		return IsValidID(index) ? Get(index) : nullptr;
	}

	/**
	 * Converts a Vehicle to SpecializedVehicle with type checking.
	 * @param v Vehicle pointer
	 * @return pointer to SpecializedVehicle
	 */
	static T *From(Vehicle *v)
	{
		assert(v->type == Type);
		return (T *)v;
	}

	/**
	 * Converts a const Vehicle to const SpecializedVehicle with type checking.
	 * @param v Vehicle pointer
	 * @return pointer to SpecializedVehicle
	 */
	static const T *From(const Vehicle *v)
	{
		assert(v->type == Type);
		return (const T *)v;
	}

	/**
	 * Update vehicle sprite- and position caches
	 * @param force_update Force updating the vehicle on the viewport.
	 * @param update_delta Also update the delta?
	 */
	void UpdateViewport(bool force_update, bool update_delta)
	{
		bool sprite_has_changed = false;

		/* Skip updating sprites on dedicated servers without screen */
		if (_network_dedicated) return;

		/* Explicitly choose method to call to prevent vtable dereference -
		 * it gives ~3% runtime improvements in games with many vehicles */
		if (update_delta) ((T *)this)->T::UpdateDeltaXY();

		/*
		 * Only check for a new sprite sequence if the vehicle direction
		 * has changed since we last checked it, assuming that otherwise
		 * there won't be enough change in bounding box or offsets to need
		 * to resolve a new sprite.
		 */
		if (this->direction != this->sprite_cache.last_direction || this->sprite_cache.is_viewport_candidate) {
			VehicleSpriteSeq seq;

			((T*)this)->T::GetImage(this->direction, EIT_ON_MAP, &seq);
			if (this->sprite_cache.sprite_seq != seq) {
				sprite_has_changed = true;
				this->sprite_cache.sprite_seq = seq;
			}

			this->sprite_cache.last_direction = this->direction;
			this->sprite_cache.revalidate_before_draw = false;
		} else {
			/*
			 * A change that could potentially invalidate the sprite has been
			 * made, signal that we should still resolve it before drawing on a
			 * viewport.
			 */
			this->sprite_cache.revalidate_before_draw = true;
		}

		if (force_update || sprite_has_changed) {
			this->Vehicle::UpdateViewport(true);
		}
	}

	/**
	 * Returns an iterable ensemble of all valid vehicles of type T
	 * @param from index of the first vehicle to consider
	 * @return an iterable ensemble of all valid vehicles of type T
	 */
	static Pool::IterateWrapper<T> Iterate(size_t from = 0) { return Pool::IterateWrapper<T>(from); }
};





/** Generates sequence of free UnitID numbers */
struct FreeUnitIDGenerator {
	bool *cache;  ///< array of occupied unit id numbers
	UnitID maxid; ///< maximum ID at the moment of constructor call
	UnitID curid; ///< last ID returned; 0 if none

	FreeUnitIDGenerator(VehicleType type, CompanyID owner);
	UnitID NextID();

	/** Releases allocated memory */
	~FreeUnitIDGenerator() { free(this->cache); }
};





bool             HandleBreakdown         (Vehicle *vehicle);
Order           *GetVehicleOrder         (const Vehicle *vehicle, int index);
StationIDStack   GetNextStoppingStation  (const Vehicle *vehicle);





#endif /* VEHICLE_BASE_H */
