/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file vehicle_func.h Functions related to vehicles. */



#ifndef VEHICLE_FUNC_H
#define VEHICLE_FUNC_H



#include "vehicle_type.h"

#include "engine_type.h"
#include "livery.h"

#include "../gfx_type.h"
#include "../direction_type.h"
#include "../command_type.h"
#include "../transport_type.h"
#include "../newgrf/newgrf_config.h"
#include "../infrastructure/track_type.h"



#define is_custom_sprite(x) (x >= 0xFD)
#define IS_CUSTOM_FIRSTHEAD_SPRITE(x) (x == 0xFD)
#define IS_CUSTOM_SECONDHEAD_SPRITE(x) (x == 0xFE)



static const int VEHICLE_PROFIT_MIN_AGE = DAYS_IN_YEAR * 2; ///< Only vehicles older than this have a meaningful profit.
static const Money VEHICLE_PROFIT_THRESHOLD = 10000;        ///< Threshold for a vehicle to be considered making good profit.



/**
 * Helper to check whether an image index is valid for a particular vehicle.
 * @tparam T The type of vehicle.
 * @param image_index The image index to check.
 * @return True iff the image index is valid.
 */
template <VehicleType T>
bool IsValidImageIndex(uint8 image_index);



typedef Vehicle *VehicleFromPosProc(Vehicle *v, void *data);
typedef std::vector<VehicleID> VehicleSet;



void VehicleServiceInDepot(Vehicle *v);
uint CountVehiclesInChain(const Vehicle *v);
void FindVehicleOnPos(TileIndex tile, void *data, VehicleFromPosProc *proc);
void FindVehicleOnPosXY(int x, int y, void *data, VehicleFromPosProc *proc);
bool HasVehicleOnPos(TileIndex tile, void *data, VehicleFromPosProc *proc);
bool HasVehicleOnPosXY(int x, int y, void *data, VehicleFromPosProc *proc);
void CallVehicleTicks();
uint8 CalcPercentVehicleFilled(const Vehicle *v, StringID *colour);

void VehicleLengthChanged(const Vehicle *u);

byte VehicleRandomBits();
void ResetVehicleHash();
void ResetVehicleColourMap();

byte GetBestFittingSubType(Vehicle *v_from, Vehicle *v_for, CargoID dest_cargo_type);

void ViewportAddVehicles(DrawPixelInfo *dpi);

void ShowNewGrfVehicleError(EngineID engine, StringID part1, StringID part2, GRFBugs bug_type, bool critical);
CommandCost TunnelBridgeIsFree(TileIndex tile, TileIndex endtile, const Vehicle *ignore = nullptr);

void DecreaseVehicleValue(Vehicle *v);
void CheckVehicleBreakdown(Vehicle *v);
void AgeVehicle(Vehicle *v);
void VehicleEnteredDepotThisTick(Vehicle *v);

UnitID GetFreeUnitNumber(VehicleType type);

void VehicleEnterDepot(Vehicle *v);
void VehicleExitDepot(Vehicle *v);

bool CanBuildVehicleInfrastructure(VehicleType type, byte subtype = 0);





/** Position information of a vehicle after it moved */

struct GetNewVehiclePosResult {

	int        x;                ///< x position of the vehicle after moving
	int        y;                ///< y position of the vehicle after moving
	TileIndex  old_tile;         ///< Current tile of the vehicle
	TileIndex  new_tile;         ///< Tile of the vehicle after moving
};





GetNewVehiclePosResult  GetNewVehiclePos     (const Vehicle *v);
Direction               GetDirectionTowards  (const Vehicle *v, int x, int y);





CommandCost  EnsureNoVehicleOnGround   (TileIndex tile);
CommandCost  EnsureNoTrainOnTrackBits  (TileIndex tile, TrackBits track_bits);





extern  VehicleID  _new_vehicle_id;
extern  uint       _returned_refit_capacity;
extern  uint16     _returned_mail_refit_capacity;





void    AutoreplaceVehicles               ();
bool    CanVehicleUseStation              (EngineID engine_type, const struct Station *station);
bool    CanVehicleUseStation              (const Vehicle *vehicle, const struct Station *station);
void    ReleaseDisastersTargetingVehicle  (VehicleID vehicle);
void    GetVehicleSet                     (VehicleSet &vehicle_set, Vehicle *vehicle, uint8 num_vehicles);
void    CheckCargoCapacity                (Vehicle *vehicle);
void    ToNextOrderIndex                  (Vehicle *vehicle);
void    AddToShared                       (Vehicle *shared_chain, Vehicle *vehicle_add);
void    RemoveFromShared                  (Vehicle *vehicle);
void    LeaveStation                      (Vehicle *vehicle);
byte    GetNumOrders                      (const Vehicle *vehicle);
bool    IsEngineCountable                 (const Vehicle *vehicle);
void    HandlePathfindingResult           (Vehicle *vehicle, bool path_found);
uint16  &GetGroundVehicleFlags            (Vehicle *vehicle);
bool    IsCompanyBuildableVehicleType     (VehicleType type);
uint32  GetCmdBuildVeh                    (VehicleType type);
uint32  GetCmdSellVeh                     (VehicleType type);
uint32  GetCmdRefitVeh                    (VehicleType type);
uint32  GetCmdSendToDepot                 (VehicleType type);





#endif /* VEHICLE_FUNC_H */
