/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file vehicle_logic.cpp Handling of vehicle logic. */





#include "../stdafx.h"

#include "vehicle_logic.h"

#include "vehicle_base.h"
#include "vehicle_func.h"
#include "aircraft.h"

#include "facilities/airport.h"
#include "facilities/depot_base.h"
#include "facilities/waypoint_base.h"

#include "orders/order.h"
#include "orders/order_func_global.h"
#include "orders/order_logic.h"
#include "orders/vehicle_orders.h"

#include "command_func.h"

#include "debug.h"





/**
 * Update the vehicle's destination tile from an order.
 * @param order the order the vehicle currently has
 * @param v the vehicle to update
 * @param conditional_depth the depth (amount of steps) to go with conditional orders,
 * this to prevent infinite loops.
 * @param pbs_look_ahead Whether we are forecasting orders for pbs reservations in advance.
 * If true, the order indices must not be modified.
 */

bool UpdateVehicleDestination (
	Vehicle *vehicle, const Order *order, int conditional_depth, bool pbs_look_ahead
) {

	if (conditional_depth > GetNumOrders (vehicle)) {

		vehicle->current_order.Free ();
		vehicle->SetDestTile        (0);

		return false;
	}

	switch (order->GetType()) {

		case OT_GOTO_STATION: {

			vehicle->SetDestTile (vehicle->GetOrderStationLocation (order->dest));

			return true;
		}

		case OT_GOTO_WAYPOINT: {

			vehicle->SetDestTile (Waypoint::Get (order->dest)->xy);

			return true;
		}

		case OT_GOTO_DEPOT: {

			if ((order->GetDepotOrderType() & ODTFB_SERVICE) && !vehicle->NeedsServicing()) {

				assert(!pbs_look_ahead);
//				ToNextOrderIndex(vehicle);

				break;
			}

			if (vehicle->current_order.GetDepotActionType() & ODATFB_NEAREST_DEPOT) {

				/* We need to search for the nearest depot (hangar). */
				TileIndex      location;
				DestinationID  destination;
				bool           reverse;

				if (vehicle->FindClosestDepot (&location, &destination, &reverse)) {

					/* PBS reservations cannot reverse */
					if (pbs_look_ahead && reverse) return false;

					vehicle->SetDestTile                 (location);
					vehicle->current_order.MakeGoToDepot (
															destination, vehicle->current_order.GetDepotOrderType(),
															vehicle->current_order.GetNonStopType(),
															(OrderDepotActionFlags)(vehicle->current_order.GetDepotActionType()
																					& ~ODATFB_NEAREST_DEPOT),
															vehicle->current_order.GetRefitCargo()
														 )
					;

					/* If there is no depot in front, reverse automatically (trains only) */
					if (vehicle->type == VEH_TRAIN && reverse) {

						CommandContainer cmdcont;
							cmdcont.tile     = vehicle->tile;
							cmdcont.p1       = vehicle->index;
							cmdcont.p2       = 0;
							cmdcont.dc_flags = DC_EXEC;
							cmdcont.cmd      = CMD_REVERSE_TRAIN_DIRECTION;

						ProcessCmd (cmdcont);
					}

					if (vehicle->type == VEH_AIRCRAFT) {

						Aircraft *aircraft = Aircraft::From (vehicle);

						if (aircraft->state == FLYING && aircraft->targetairport != destination) {

							/* The aircraft is now heading for a different hangar than the next in the orders */
							extern void AircraftNextAirportPos_and_Order (Aircraft *aircraft);

							AircraftNextAirportPos_and_Order (aircraft);
						}
					}

					return true;
				}

				/* If there is no depot, we cannot help PBS either. */
				if (pbs_look_ahead) return false;

//				ToNextOrderIndex(vehicle);

			} else {

				if (vehicle->type != VEH_AIRCRAFT) {

					vehicle->SetDestTile(Depot::Get (order->dest)->xy);

				} else {

					Aircraft       *aircraft    = Aircraft::From (vehicle);
					DestinationID   destination = aircraft->current_order.dest;

					if (aircraft->targetairport != destination) {

						/* The aircraft is now heading for a different hangar than the next in the orders */
						aircraft->SetDestTile (aircraft->GetOrderStationLocation (destination));
					}
				}

				return true;
			}

			break;
		}

		case OT_CONDITIONAL: {

			assert(!pbs_look_ahead);

			VehicleOrderID next_order = ProcessConditionalOrder (order, vehicle);

			if (next_order != INVALID_VEH_ORDER_ID) {

				/* Jump to next_order. cur_implicit_order_index becomes exactly that order,
				 * current_order_index might come after next_order. */
				vehicle->current_order_index  = next_order;
				vehicle->current_order_time  += GetVehicleOrder (vehicle, vehicle->current_order_index)->travel_time;

				/* Disable creation of implicit orders.
				 * When inserting them we do not know that we would have to make the conditional orders point to them. */
				if (vehicle->IsGroundVehicle()) {

					uint16 &gv_flags = GetGroundVehicleFlags (vehicle);
				}

			} else {

				ToNextOrderIndex (vehicle);
			}

			break;
		}

		default: {

			vehicle->SetDestTile (0);

			return false;
		}
	}

	assert (vehicle->current_order_index < GetNumOrders (vehicle));

	/* Get the current order */
	order = GetVehicleOrder (vehicle, vehicle->current_order_index);

	if (order == nullptr) {

		vehicle->current_order.Free ();
		vehicle->SetDestTile        (0);

		return false;
	}

	vehicle->current_order = *order;

	return UpdateVehicleDestination (vehicle, order, conditional_depth + 1, pbs_look_ahead);
}





/**
 * Clamp the service interval to the correct min/max. The actual min/max values
 * depend on whether it's in percent or days.
 * @param interval proposed service interval
 * @return Clamped service interval
 */

uint16 GetServiceIntervalClamped (
	uint interval, bool ispercent
) {

	return ispercent
			? Clamp (interval, MIN_SERVINT_PERCENT, MAX_SERVINT_PERCENT)
			: Clamp (interval, MIN_SERVINT_DAYS,    MAX_SERVINT_DAYS)
	;
}





/**
 * Check if an aircraft has enough range for an order list distance.
 * @param v_new Aircraft to check.
 * @param v_order Vehicle currently holding the order list.
 * @param first First order in the source order list.
 * @return True if the aircraft has enough range for the orders, false otherwise.
 */

bool CheckAircraftOrderDistance (
	const Aircraft *v_new, const Vehicle *v_order, const Order *first
) {

	if (first == nullptr || v_new->acache.cached_max_range == 0) {

		return true;
	}

	/* Iterate over all orders to check the distance between all
	 * 'goto' orders and their respective next order (of any type). */
	for (const Order *o = first; o != nullptr; o = o->next) {

		switch (o->GetType()) {

			case OT_GOTO_STATION:
			case OT_GOTO_DEPOT:
			case OT_GOTO_WAYPOINT: {

				/* If we don't have a next order, we've reached the end and must check the first order instead. */
				if (GetOrderDistance (o,
									  o->next != nullptr ? o->next : first,
									  v_order
									 )
					> v_new->acache.cached_max_range_sqr
				) {

					return false;
				}

				break;
			}

			default: {

				break;
			}
		}
	}

	return true;
}
