/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file vehicle_loop.cpp Unified handling of all vehicles. */





#include "vehicle_loop.h"

#include "vehicle_base.h"
#include "vehicle_func.h"
#include "vehicle_gui.h"
#include "aircraft.h"
#include "roadveh.h"
#include "ship.h"
#include "train.h"
#include "effectvehicle_base.h"
#include "effectvehicle_func.h"
#include "vehicle_logic.h"

#include "../economy/economy_base.h"

#include "../facilities/station_base.h"
#include "../facilities/depot_map.h"

#include "../handling/loadunload_func.h"
//#include "../linkgraph/refresh.h"

#include "../newgrf/newgrf_sound.h"
#include "../newgrf/newgrf_station.h"

#include "../orders/order_func_global.h"
#include "../sound_func.h"

#include "../debug.h"





/**
 * Unified handling of vehicles.
*/

int HandleVehicle (
	Vehicle *vehicle
) {

	switch (vehicle->type) {

		/* Player vehicles */
		case VEH_AIRCRAFT:
		case VEH_ROAD:
		case VEH_SHIP:
		case VEH_TRAIN: {

			if (vehicle->vehstatus & VS_CRASHED) {

				return HandleVehicleCrash (vehicle);
			}

			if (vehicle->vehstatus & VS_STOPPED) {

				return false;
			}

			if (HandleBreakdown (vehicle)) {

				return false;
			}

			if (vehicle->vehicle_location & VL_AT_NEW_TILE) {

				VehicleAtNewTile (vehicle);
			}

			if (vehicle->vehicle_location & VL_SKIP_LOCATION) {

				vehicle->vehicle_location &= ~VL_SKIP_LOCATION;

				VehicleToNextDestination (vehicle);
			}

			/* Speed doesn't drop below 5 because of an issue with realistic acceleration? */
			if (vehicle->vehicle_location & VL_ENTER_WAYPOINT) {

				if (vehicle->cur_speed <= 5) {

					vehicle->vehicle_location |=  VL_AT_WAYPOINT;
					vehicle->vehicle_action   |=  VA_SET_WAITING;
					vehicle->vehicle_location &= ~VL_ENTER_WAYPOINT;

					UpdateVehicleTimetable (vehicle, ARRIVAL);

					return true;
				}
			}

			if (vehicle->vehicle_action & VA_LEAVE) {

				HandleVehicleLeaving (vehicle);

				return true;
			}

			if (vehicle->vehicle_action & VA_SET_WAITING) {

				SetVehicleWaitingTime (vehicle);
			}

			if (vehicle->vehstatus & VS_WAITING) {

				if (HandleVehicleWaiting (vehicle)) {

					return true;
				}
			}

			if (
				   (vehicle->vehstatus & VS_UNLOADING)
				|| (vehicle->vehstatus & VS_LOADING)
			) {

				return true;
			}

			break;
		}

		/* Other vehicles */
		default: {

			if (vehicle->vehstatus & VS_CRASHED) {

				return HandleVehicleCrash (vehicle);
			}

			break;
		}

	} /* end switch */

	/* Function did not apply. */
	return neither;
}





/**
 * Handle crash logic for the vehicle.
 */

bool HandleVehicleCrash (
	Vehicle *vehicle
) {

	switch (vehicle->type) {

		case VEH_AIRCRAFT: {

			return HandleCrashedAircraft ((Aircraft *)vehicle);
		}

		case VEH_ROAD: {

			return RoadVehIsCrashed ((RoadVehicle *)vehicle);
		}

		case VEH_SHIP: {

			return true;
		}

		case VEH_TRAIN: {

			return HandleCrashedTrain ((Train *)vehicle);
		}

		default: {

			return false;
		}
	}
}





/**
 * Handle all of the aspects of a vehicle breakdown
 * This includes adding smoke and sounds, and ending the breakdown when appropriate.
 * @return true iff the vehicle is stopped because of a breakdown
 * @note This function always returns false for aircraft, since these never stop for breakdowns
 */

bool HandleBreakdown (
	Vehicle *vehicle
) {

	/*
	 * Possible states for Vehicle::breakdown_ctr
	 *
	 *  0 - vehicle is running normally
	 *  1 - vehicle is currently broken down
	 *  2 - vehicle is going to break down now
	 * >2 - vehicle is counting down to the actual breakdown event
	 */

	switch (vehicle->breakdown_ctr) {

		case 0: {

			return false;
		}

		case 2: {

			vehicle->breakdown_ctr = 1;

			if (vehicle->breakdowns_since_last_service != 255) {

				vehicle->breakdowns_since_last_service++;
			}

			if (vehicle->type == VEH_AIRCRAFT) {

				/* Aircraft just need vehicle flag, the rest is handled elsewhere */
				vehicle->vehstatus |= VS_AIRCRAFT_BROKEN;

			} else {

				vehicle->cur_speed = 0;

				if (!PlayVehicleSound (vehicle, VSE_BREAKDOWN)) {

					bool train_or_ship = vehicle->type == VEH_TRAIN || vehicle->type == VEH_SHIP;

					SndPlayVehicleFx (
										(_settings_game.game_creation.landscape != LT_TOYLAND)
										? (train_or_ship ? SND_10_BREAKDOWN_TRAIN_SHIP         : SND_0F_BREAKDOWN_ROADVEHICLE)
										: (train_or_ship ? SND_3A_BREAKDOWN_TRAIN_SHIP_TOYLAND : SND_35_BREAKDOWN_ROADVEHICLE_TOYLAND)
									 , vehicle
					);
				}

				if (!(vehicle->vehstatus & VS_HIDDEN) && !HasBit(EngInfo(vehicle->engine_type)->misc_flags, EF_NO_BREAKDOWN_SMOKE)) {

					EffectVehicle *u = CreateEffectVehicleRel(vehicle, 4, 4, 5, EV_BREAKDOWN_SMOKE);

					if (u != nullptr) {

						u->animation_state = (uint16)vehicle->breakdown_delay * 2;
					}
				}
			}

			vehicle->MarkDirty(); // Update graphics after speed is zeroed

			SetWindowDirty (WC_VEHICLE_VIEW,    vehicle->index);
			SetWindowDirty (WC_VEHICLE_DETAILS, vehicle->index);

			FALLTHROUGH;
		}

		case 1: {

			/* Aircraft breakdowns end only when arriving at the airport */
			if (vehicle->type == VEH_AIRCRAFT) return false;

			/* For trains vehicle function is called twice per tick, so decrease v->breakdown_delay at half the rate */
			if ((vehicle->tick_counter & (vehicle->type == VEH_TRAIN ? 3 : 1)) == 0) {

				if (--vehicle->breakdown_delay == 0) {

					vehicle->breakdown_ctr = 0;
					vehicle->MarkDirty();

					SetWindowDirty (WC_VEHICLE_VIEW, vehicle->index);
				}
			}

			return true;
		}

		default: {

			if (!vehicle->current_order.IsType(OT_LOADING)) {

				vehicle->breakdown_ctr--;
			}

			return false;
		}
	}
}





void VehicleAtNewTile (
	Vehicle *vehicle
) {

	if (IsTileType (vehicle->tile, MP_STATION)) {

	// todo: add skip if tile is not station/waypoint/depot
	if (   !(vehicle->vehicle_location & VL_AT_STATION)
		|| !(vehicle->vehicle_location & VL_ENTER_WAYPOINT)
		|| !(vehicle->vehicle_location & VL_ENTER_DEPOT)
	) {

		ShouldStopAtStation (vehicle);
	}

	switch (vehicle->type) {

		case VEH_TRAIN: {

			CheckNextTrainTile ((Train *)vehicle);

			break;
		}

		default: {

			break;
		}
	}

	vehicle->vehicle_location &= ~VL_AT_NEW_TILE;
	}
}





void VehicleToNextDestination (
	Vehicle *vehicle
) {

	ToNextOrderIndex         (vehicle);

	Order *order           = GetVehicleOrder (vehicle, vehicle->current_order_index);
	vehicle->current_order = *order;

	InvalidateVehicleOrder   (vehicle, VIWD_MODIFY_ORDERS);
	UpdateVehicleDestination (vehicle, GetVehicleOrder (vehicle, vehicle->current_order_index) );
}





/**
 * Set the waiting state and time for the vehicle.
 */

void SetVehicleWaitingTime (
	Vehicle *vehicle
) {

	if (vehicle->current_order.wait_time) {

		vehicle->vehstatus      |=  VS_WAITING;
		vehicle->vehicle_action &= ~VA_SET_WAITING;

		vehicle->current_order.scheduled_departure = _time_tick + vehicle->current_order.wait_time;
	}
}



/**
 * Handle the waiting time for a vehicle.
 */

bool HandleVehicleWaiting (
	Vehicle *vehicle
) {

	if (_time_tick < vehicle->current_order.scheduled_departure) {

		vehicle->vehstatus      |=  VS_WAITING;

		return true;

	} else {

		vehicle->vehstatus      &= ~VS_WAITING;
		vehicle->vehicle_action |=  VA_LEAVE;
	}

	return false;
}





/**
 * Handle logic for the vehicle leaving a facility.
 */

void HandleVehicleLeaving (
	Vehicle *vehicle
) {

	if (
			(vehicle->vehicle_location & VL_AT_STATION)
		&& !(vehicle->vehicle_location & VL_AT_BAY)
	) {

		vehicle->vehicle_location |=  VL_LEAVE_STATION;
		vehicle->vehicle_action   &= ~VA_LEAVE;

		LeaveStation (vehicle);

		UpdateVehicleTimetable   (vehicle, DEPARTURE);
		VehicleToNextDestination (vehicle);

		return;
	}

	if (vehicle->vehicle_location & VL_AT_WAYPOINT) {

		vehicle->vehicle_location &= ~VL_AT_WAYPOINT;
		vehicle->vehicle_action   &= ~VA_LEAVE;

		UpdateVehicleTimetable   (vehicle, DEPARTURE);
		VehicleToNextDestination (vehicle);

		if (vehicle->type == VEH_TRAIN) {

			((Train *)vehicle)->force_proceed  = TFP_NONE;
			vehicle           ->vehicle_flags |= VF_COULD_REVERSE;

			return;
		}
	}

	if (vehicle->vehicle_location & VL_AT_DEPOT) {

		vehicle->vehicle_location |=  VL_LEAVE_DEPOT;
		vehicle->vehicle_action   &= ~VA_LEAVE;

		return;
	}

	return;
}





/**
 * Perform all actions when leaving a station.
 * @pre this->current_order.IsType(OT_LOADING)
 */

void LeaveStation (
	Vehicle *vehicle
) {

	if (vehicle->cargo_payment != nullptr) {

		delete vehicle->cargo_payment;
	}

	if (
		!(vehicle->current_order.GetLoadType() & OLFB_NO_LOAD)
		|| !(vehicle->current_order.GetUnloadType() & OUFB_NO_UNLOAD)
	) {

		const Order *curr_order = &vehicle->current_order;

		if (CanLeaveWithCargo (curr_order, vehicle->last_loading_station != INVALID_STATION)) {

			/* Refresh next hop stats to make sure we've done that at least once
			 * during the stop and that refit_cap == cargo_cap for each vehicle in
			 * the consist. */
			vehicle->ResetRefitCaps();
//			LinkRefresher::Run (vehicle);

			/* if the vehicle could load here or could stop with cargo loaded set the last loading station */
			vehicle->last_loading_station = vehicle->last_station_visited;

		} else {

			/* if the vehicle couldn't load and had to unload or transfer everything
			 * set the last loading station to invalid as it will leave empty. */
			vehicle->last_loading_station = INVALID_STATION;
		}
	}

	Station *station = Station::Get (vehicle->last_station_visited);
	CancelReservation (vehicle, INVALID_STATION, station);

	if (vehicle->vehstatus & VS_STATION_LISTED) {

		station->loading_vehicles.remove (vehicle);
	}

	HideFillingPercent (&vehicle->fill_percent_te_id);
	vehicle->trip_occupancy = CalcPercentVehicleFilled (vehicle, nullptr);

	if (vehicle->type == VEH_TRAIN && !(vehicle->vehstatus & VS_CRASHED)) {

		/* Trigger station animation (trains only) */
		if (IsTileType (vehicle->tile, MP_STATION)) {

			TriggerStationRandomisation (station, vehicle->tile, SRT_TRAIN_DEPARTS);
			TriggerStationAnimation     (station, vehicle->tile, SAT_TRAIN_DEPARTS);
		}

		SetBit ( ((Train *)vehicle)->flags, VRF_LEAVING_STATION);
	}

	if (vehicle->type == VEH_TRAIN) {

		((Train *)vehicle)->force_proceed = TFP_NONE;
	}

	vehicle->vehstatus        &= ~VS_WAITING;
	vehicle->vehicle_location &= ~VL_AT_STATION;
	vehicle->vehicle_location &= ~VL_LEAVE_STATION;

	if (vehicle->type == VEH_TRAIN) {

		vehicle->vehicle_flags |= VF_COULD_REVERSE;
	}

	vehicle->PlayLeaveStationSound();
	vehicle->MarkDirty();
}
