/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file vehicle_type.h Types related to vehicles. */





#ifndef VEHICLE_TYPE_H
#define VEHICLE_TYPE_H





#include "../core/enum_type.hpp"
#include "../core/pool_type.hpp"





/** The type all our vehicle IDs have. */
typedef uint32 VehicleID;





static const  int        GROUND_ACCELERATION            = 9800;        ///< Acceleration due to gravity, 9.8 m/s^2
static const  int32      INVALID_COORD                  = 0x7fffffff;  ///< Sentinel for an invalid coordinate
static const  VehicleID  INVALID_VEHICLE                = 0xFFFFF;     ///< Constant representing a non-existing vehicle.
static const  uint       VEHICLE_LENGTH                 = 8;           ///< The length of a vehicle in tile units
static const  uint       MAX_LENGTH_VEHICLE_NAME_CHARS  = 32;          ///< The maximum length of a vehicle name
																	   ///< in characters including '\0'




/** Available vehicle types. It needs to be 8bits, because we save and load it as such */

enum VehicleType : byte {

	VEH_BEGIN,

	VEH_TRAIN = VEH_BEGIN,        ///< %Train vehicle type.
	VEH_ROAD,                     ///< Road vehicle type.
	VEH_SHIP,                     ///< %Ship vehicle type.
	VEH_AIRCRAFT,                 ///< %Aircraft vehicle type.

	VEH_COMPANY_END,              ///< Last company-ownable type.

	VEH_EFFECT = VEH_COMPANY_END, ///< Effect vehicle type (smoke, explosions, sparks, bubbles)
	VEH_DISASTER,                 ///< Disaster vehicle type.

	VEH_END,
	VEH_INVALID = 0xFF,           ///< Non-existing type of vehicle.
};

DECLARE_POSTFIX_INCREMENT(VehicleType)

/** Helper information for extract tool. */
template <> struct EnumPropsT<VehicleType> : MakeEnumPropsT<VehicleType, byte, VEH_TRAIN, VEH_END, VEH_INVALID, 3> {};





/** Pathfinding option states */

enum VehiclePathFinders {

	// Original PathFinder (OPF) used to be 0
	VPF_NPF  = 1, ///< New PathFinder
	VPF_YAPF = 2, ///< Yet Another PathFinder
};



/** Flags to add to p1 for goto depot commands. */

enum DepotCommand {

	DEPOT_SERVICE       = (1U << 28), ///< The vehicle will leave the depot right after arrival (service only)
	DEPOT_MASS_SEND     = (1U << 29), ///< Tells that it's a mass send to depot command (type in VLW flag)
	DEPOT_DONT_CANCEL   = (1U << 30), ///< Don't cancel current goto depot command if any
	DEPOT_LOCATE_HANGAR = (1U << 31), ///< Find another airport if the target one lacks a hangar
	DEPOT_COMMAND_MASK  = 0xFU << 28,
};





/** Vehicle acceleration models. */

enum AccelerationModel {

	AM_ORIGINAL,
	AM_REALISTIC,
};



/** Visualisation contexts of vehicles and engines. */

enum EngineImageType {

	EIT_ON_MAP     = 0x00,  ///< Vehicle drawn in viewport.
	EIT_IN_DEPOT   = 0x10,  ///< Vehicle drawn in depot.
	EIT_IN_DETAILS = 0x11,  ///< Vehicle drawn in vehicle details, refit window, ...
	EIT_IN_LIST    = 0x12,  ///< Vehicle drawn in vehicle list, group list, ...
	EIT_PURCHASE   = 0x20,  ///< Vehicle drawn in purchase list, autoreplace gui, ...
	EIT_PREVIEW    = 0x21,  ///< Vehicle drawn in preview window, news, ...
};





/** Vehicle status bits in #Vehicle::vehstatus, e.g. vehicle "is". */

enum VehStatus {

	VS_HIDDEN           = 1 <<  0,  ///< Vehicle is not visible.
	VS_STOPPED          = 1 <<  1,  ///< Vehicle is stopped by the player.
	VS_UNCLICKABLE      = 1 <<  2,  ///< Vehicle is not clickable by the user (shadow vehicles).
	VS_DEFPAL           = 1 <<  3,  ///< Use default vehicle palette. @see DoDrawVehicle
	VS_SHADOW           = 1 <<  4,  ///< Vehicle is a shadow vehicle.
	VS_AIRCRAFT_BROKEN  = 1 <<  5,  ///< Aircraft is broken down.
	VS_CRASHED          = 1 <<  6,  ///< Vehicle is crashed.
	VS_WAITING          = 1 <<  7,  ///< Vehicle is waiting.
	VS_LOADING          = 1 <<  8,  ///< Vehicle is loading.
	VS_UNLOADING        = 1 <<  9,  ///< Vehicle is unloading.
	VS_TRAVELLING       = 1 << 10,  ///< Vehicle is travelling.
	VS_HALTED           = 1 << 11,  ///< Vehicle is temporarily stopped.
	VS_DEFECT           = 1 << 12,  ///< Vehicle is defective.
	VS_SERVICING        = 1 << 13,  ///< Vehicle is in service.
	VS_PARKED           = 1 << 14,  ///< Vehicle is parked.
	VS_STATION_LISTED   = 1 << 15   ///< Vehicle is on station loading list.
};





/* Action flags for a vehicle */

enum VehicleAction {

	VA_SET_WAITING      = 1 <<  0, ///< Set waiting time for vehicle
	VA_BRAKING          = 1 <<  1, ///< Vehicle is braking.
	VA_REVERSE          = 1 <<  2, ///< Reverse vehicle
	VA_LEAVE            = 1 <<  3
};





/* Location flags for a vehicle */

enum VehicleLocation {

	VL_AT_NEW_TILE      = 1 <<  0,  ///< Vehicle has enetered a new tile.
	VL_SKIP_LOCATION    = 1 <<  1,  ///< Vehicle skips current location

	VL_ENTER_STATION    = 1 <<  2,  ///< Vehicle is entering a station.
	VL_AT_STATION       = 1 <<  3,  ///< Vehicle is at a station.
	VL_LEAVE_STATION    = 1 <<  4,  ///< Vehicle is leaving a station.

	VL_ENTER_WAYPOINT   = 1 <<  5,  ///< Vehicle is entering a waypoint.
	VL_AT_WAYPOINT      = 1 <<  6,  ///< Vehicle is at a waypoint.
	VL_LEAVE_WAYPOINT   = 1 <<  7,  ///< Vehicle is leaving a waypoint.

	VL_ENTER_DEPOT      = 1 <<  8,  ///< Vehicle is entering a depot.
	VL_AT_DEPOT         = 1 <<  9,  ///< Vehicle is at a depot.
	VL_LEAVE_DEPOT      = 1 << 10,  ///< Vehicle is leaving a depot.

	VL_AT_BAY           = 1 << 11,  ///< Road vehicle specific location
	VL_AT_BAY1          = 1 << 12,
	VL_AT_BAY2          = 1 << 13
};





/** Bit numbers in #Vehicle::vehicle_flags, e.g. vehicle "has". */

enum VehicleFlags {

	VF_LOADING_FINISHED        =       0, ///< Vehicle has finished loading.
	VF_CARGO_UNLOADING         = 1 <<  0, ///< Vehicle is unloading cargo.
	VF_BUILT_AS_PROTOTYPE      = 1 <<  1, ///< Vehicle is a prototype (accepted as exclusive preview).
	VF_TIMETABLE_STARTED       = 1 <<  2, ///< Whether the vehicle has started running on the timetable yet.
	VF_AUTOFILL_TIMETABLE      = 1 <<  3, ///< Whether the vehicle should fill in the timetable automatically.
	VF_AUTOFILL_PRES_WAIT_TIME = 1 <<  4, ///< Whether non-destructive auto-fill should preserve waiting times
	VF_STOP_LOADING            = 1 <<  5, ///< Don't load anymore during the next load cycle.
	VF_PATHFINDER_LOST         = 1 <<  6, ///< Vehicle's pathfinder is lost.
	VF_SERVINT_IS_CUSTOM       = 1 <<  7, ///< Service interval is custom.
	VF_SERVINT_IS_PERCENT      = 1 <<  8, ///< Service interval is percent.
	VF_SHARED_ORDERS           = 1 <<  9, ///< Vehicle shares orders with other vehicle(s).
	VF_COULD_REVERSE           = 1 << 10, ///< Vehicle could reverse at this point.
};





/** Bit numbers used to indicate which of the #NewGRFCache values are valid. */

enum NewGRFCacheValidValues {

	NCVV_POSITION_CONSIST_LENGTH   = 0, ///< This bit will be set if the NewGRF var 40 currently stored is valid.
	NCVV_POSITION_SAME_ID_LENGTH   = 1, ///< This bit will be set if the NewGRF var 41 currently stored is valid.
	NCVV_CONSIST_CARGO_INFORMATION = 2, ///< This bit will be set if the NewGRF var 42 currently stored is valid.
	NCVV_COMPANY_INFORMATION       = 3, ///< This bit will be set if the NewGRF var 43 currently stored is valid.
	NCVV_POSITION_IN_VEHICLE       = 4, ///< This bit will be set if the NewGRF var 4D currently stored is valid.

	NCVV_END,                           ///< End of the bits.
};





/** Meaning of the various bits of the visual effect. */

enum VisualEffect {

	VE_OFFSET_START        = 0,                 ///< First bit that contains the offset (0 = front, 8 = centre, 15 = rear)
	VE_OFFSET_COUNT        = 4,                 ///< Number of bits used for the offset
	VE_OFFSET_CENTRE       = 8,                 ///< Value of offset corresponding to a position above the centre of the vehicle

	VE_TYPE_START          = 4,                 ///< First bit used for the type of effect
	VE_TYPE_COUNT          = 2,                 ///< Number of bits used for the effect type
	VE_TYPE_DEFAULT        = 0,                 ///< Use default from engine class
	VE_TYPE_STEAM          = 1,                 ///< Steam plumes
	VE_TYPE_DIESEL         = 2,                 ///< Diesel fumes
	VE_TYPE_ELECTRIC       = 3,                 ///< Electric sparks

	VE_DISABLE_EFFECT      = 6,                 ///< Flag to disable visual effect
	VE_ADVANCED_EFFECT     = VE_DISABLE_EFFECT, ///< Flag for advanced effects
	VE_DISABLE_WAGON_POWER = 7,                 ///< Flag to disable wagon power

	VE_DEFAULT = 0xFF,          ///< Default value to indicate that visual effect should be based on engine class
};





/** Models for spawning visual effects. */

enum VisualEffectSpawnModel {

	VESM_NONE              = 0, ///< No visual effect
	VESM_STEAM,                 ///< Steam model
	VESM_DIESEL,                ///< Diesel model
	VESM_ELECTRIC,              ///< Electric model

	VESM_END
};





/**
 * Enum to handle ground vehicle subtypes.
 * This is defined here instead of at #GroundVehicle because some common function require access to these flags.
 * Do not access it directly unless you have to. Use the subtype access functions.
 */

enum GroundVehicleSubtypeFlags {

	GVSF_FRONT            = 0, ///< Leading engine of a consist.
	GVSF_ARTICULATED_PART = 1, ///< Articulated part of an engine.
	GVSF_WAGON            = 2, ///< Wagon (not used for road vehicles).
	GVSF_ENGINE           = 3, ///< Engine that can be front engine, but might be placed behind another engine
							   ///< (not used for road vehicles).
	GVSF_FREE_WAGON       = 4, ///< First in a wagon chain (in depot) (not used for road vehicles).
	GVSF_MULTIHEADED      = 5, ///< Engine is multiheaded (not used for road vehicles).
};





/** Base vehicle class. */

struct BaseVehicle
{
	VehicleType type; ///< Type of vehicle
};





struct Vehicle;

struct Aircraft;
struct RoadVehicle;
struct Ship;
// struct Train;

struct EffectVehicle;
struct DisasterVehicle;





/** A vehicle pool for a little over 1 million vehicles. */
typedef Pool<Vehicle, VehicleID, 512, 0xFF000> VehiclePool;
extern VehiclePool _vehicle_pool;





/* Some declarations of functions, so we can make them friendly */
struct GroundVehicleCache;
struct LoadgameState;
extern bool LoadOldVehicle(LoadgameState *ls, int num);
extern void FixOldVehicles();

struct GRFFile;





/** Cached often queried (NewGRF) values */

struct NewGRFCache {

	/* Values calculated when they are requested for the first time after invalidating the NewGRF cache. */
	uint32 position_consist_length;   ///< Cache for NewGRF var 40.
	uint32 position_same_id_length;   ///< Cache for NewGRF var 41.
	uint32 consist_cargo_information; ///< Cache for NewGRF var 42. (Note: The cargotype is untranslated in the cache
									  ///<  because the accessing GRF is yet unknown.)
	uint32 company_information;       ///< Cache for NewGRF var 43.
	uint32 position_in_vehicle;       ///< Cache for NewGRF var 4D.
	uint8  cache_valid;               ///< Bitset that indicates which cache values are valid.
};





#endif /* VEHICLE_TYPE_H */
